"""Argparser module for ZMQ runtimes"""
import os

from ...helper import add_arg_group
from .... import helper


def mixin_zmq_runtime_parser(parser):
    """Mixing in arguments required by :class:`ZMQRuntime` into the given parser.
    :param parser: the parser instance to which we add arguments
    """

    gp = add_arg_group(parser, title='ZMQRuntime')
    gp.add_argument(
        '--port-ctrl',
        type=int,
        default=os.environ.get('JINA_CONTROL_PORT', helper.random_port()),
        help='The port for controlling the runtime, default a random port between [49152, 65535]',
    )
    gp.add_argument(
        '--ctrl-with-ipc',
        action='store_true',
        default=False,
        help='If set, use ipc protocol for control socket',
    )
    gp.add_argument(
        '--timeout-ctrl',
        type=int,
        default=5000,
        help='The timeout in milliseconds of the control request, -1 for waiting forever',
    )

    gp.add_argument(
        '--ssh-server',
        type=str,
        default=None,
        help='The SSH server through which the tunnel will be created, '
        'can actually be a fully specified `user@server:port` ssh url.',
    )
    gp.add_argument(
        '--ssh-keyfile',
        type=str,
        default=None,
        help='This specifies a key to be used in ssh login, default None. '
        'regular default ssh keys will be used without specifying this argument.',
    )
    gp.add_argument(
        '--ssh-password',
        type=str,
        default=None,
        help='The ssh password to the ssh server.',
    )



__copyright__ = "Copyright (c) 2020 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

import inspect
from functools import wraps
from typing import (
    Any,
    Dict,
    Callable,
    Tuple,
    Optional,
    Sequence,
    Iterable,
    List,
    Union,
)

from google.protobuf.struct_pb2 import Struct

from ..enums import OnErrorStrategy
from ..excepts import LengthMismatchException
from ..executors.compound import CompoundExecutor
from ..executors.decorators import wrap_func
from ..helper import (
    convert_tuple_to_list,
    cached_property,
    find_request_binding,
    _canonical_request_name,
)
from ..jaml import JAMLCompatible
from ..types.querylang import QueryLang
from ..types.sets import DocumentSet

# noinspection PyUnreachableCode
if False:
    # fix type-hint complain for sphinx and flake
    from ..peapods.runtimes.zmq.zed import ZEDRuntime
    from ..executors import AnyExecutor
    from ..logging.logger import JinaLogger
    from ..types.message import Message
    from ..types.request import Request
    from ..types.sets import QueryLangSet
    from ..types.document import Document


def store_init_kwargs(func: Callable) -> Callable:
    """Mark the args and kwargs of :func:`__init__` later to be stored via :func:`save_config` in YAML

    :param func: the Callable to wrap
    :return: the wrapped Callable
    """

    @wraps(func)
    def _arg_wrapper(self, *args, **kwargs):
        if func.__name__ != '__init__':
            raise TypeError(
                'this decorator should only be used on __init__ method of a driver'
            )
        taboo = {'self', 'args', 'kwargs'}
        all_pars = inspect.signature(func).parameters
        tmp = {k: v.default for k, v in all_pars.items() if k not in taboo}
        tmp_list = [k for k in all_pars.keys() if k not in taboo]
        # set args by aligning tmp_list with arg values
        for k, v in zip(tmp_list, args):
            tmp[k] = v
        # set kwargs
        for k, v in kwargs.items():
            if k in tmp:
                tmp[k] = v

        if self.store_args_kwargs:
            if args:
                tmp['args'] = args
            if kwargs:
                tmp['kwargs'] = {k: v for k, v in kwargs.items() if k not in taboo}

        if hasattr(self, '_init_kwargs_dict'):
            self._init_kwargs_dict.update(tmp)
        else:
            self._init_kwargs_dict = tmp
        convert_tuple_to_list(self._init_kwargs_dict)
        f = func(self, *args, **kwargs)
        return f

    return _arg_wrapper


class QuerySetReader:
    """
    :class:`QuerySetReader` allows a driver to read arguments from the protobuf message. This allows a
    driver to override its behavior based on the message it receives. Extremely useful in production, for example,
    get ``top_k`` results, doing pagination, filtering.

    To register the field you want to read from the message, simply register them in :meth:`__init__`.
    For example, ``__init__(self, arg1, arg2, **kwargs)`` will allow the driver to read field ``arg1`` and ``arg2`` from
    the message. When they are not found in the message, the value ``_arg1`` and ``_arg2`` will be used. Note the underscore
    prefix.

    .. note::
        - To set default value of ``arg1``, use ``self._arg1 =``, note the underscore in the front.
        - To access ``arg1``, simply use ``self.arg1``. It automatically switch between default ``_arg1`` and ``arg1`` from the request.

    For successful value reading, the following condition must be met:

        - the ``name`` in the proto must match with the current class name
        - the ``disabled`` field in the proto should not be ``False``
        - the ``priority`` in the proto should be strictly greater than the driver's priority (by default is 0)
        - the field name must exist in proto's ``parameters``

    .. warning::
        For the sake of cooperative multiple inheritance, do NOT implement :meth:`__init__` for this class
    """

    @property
    def as_querylang(self):
        """Render as QueryLang parameters.


        .. # noqa: DAR201"""
        parameters = {
            name: getattr(self, name) for name in self._init_kwargs_dict.keys()
        }
        return QueryLang(
            {
                'name': self.__class__.__name__,
                'priority': self._priority,
                'parameters': parameters,
            }
        )

    def _get_parameter(self, key: str, default: Any):
        if getattr(self, 'queryset', None):
            for q in self.queryset:
                if (
                    not q.disabled
                    and self.__class__.__name__ == q.name
                    and q.priority > self._priority
                    and key in q.parameters
                ):
                    ret = q.parameters[key]
                    return dict(ret) if isinstance(ret, Struct) else ret
        return getattr(self, f'_{key}', default)

    def __getattr__(self, name: str):
        # https://docs.python.org/3/reference/datamodel.html#object.__getattr__
        if name == '_init_kwargs_dict':
            # raise attribute error to avoid recursive call
            raise AttributeError
        if name in self._init_kwargs_dict:
            return self._get_parameter(name, default=self._init_kwargs_dict[name])
        raise AttributeError


class DriverType(type(JAMLCompatible), type):
    """A meta class representing a Driver

    When a new Driver is created, it gets registered
    """

    def __new__(cls, *args, **kwargs):
        """Create and register a new class with this meta class.

        :param args: additional positional arguments which are just used for the parent initialization
        :param kwargs: additional key value arguments which are just used for the parent initialization
        :return: the newly registered class
        """
        _cls = super().__new__(cls, *args, **kwargs)
        return cls.register_class(_cls)

    @staticmethod
    def register_class(cls):
        """Register a class

        :param cls: the class
        :return: the class, after being registered
        """
        reg_cls_set = getattr(cls, '_registered_class', set())
        if cls.__name__ not in reg_cls_set or getattr(cls, 'force_register', False):
            wrap_func(cls, ['__init__'], store_init_kwargs)
            # wrap_func(cls, ['__call__'], as_reduce_method)

            reg_cls_set.add(cls.__name__)
            setattr(cls, '_registered_class', reg_cls_set)
        return cls


class BaseDriver(JAMLCompatible, metaclass=DriverType):
    """A :class:`BaseDriver` is a logic unit above the :class:`jina.peapods.runtimes.zmq.zed.ZEDRuntime`.
    It reads the protobuf message, extracts/modifies the required information and then return
    the message back to :class:`jina.peapods.runtimes.zmq.zed.ZEDRuntime`.

    A :class:`BaseDriver` needs to be :attr:`attached` to a :class:`jina.peapods.runtimes.zmq.zed.ZEDRuntime` before
    using. This is done by :func:`attach`. Note that a deserialized :class:`BaseDriver` from file is always unattached.

    :param priority: the priority of its default arg values (hardcoded in Python). If the
         received ``QueryLang`` has a higher priority, it will override the hardcoded value
    :param args: not used (kept to maintain interface)
    :param kwargs: not used (kept to maintain interface)
    """

    store_args_kwargs = False  #: set this to ``True`` to save ``args`` (in a list) and ``kwargs`` (in a map) in YAML config

    def __init__(self, priority: int = 0, *args, **kwargs):
        self.attached = False  # : represent if this driver is attached to a
        # :class:`jina.peapods.runtimes.zmq.zed.ZEDRuntime` (& :class:`jina.executors.BaseExecutor`)
        self.runtime = None  # type: Optional['ZEDRuntime']
        self._priority = priority

    def attach(self, runtime: 'ZEDRuntime', *args, **kwargs) -> None:
        """Attach this driver to a :class:`jina.peapods.runtimes.zmq.zed.ZEDRuntime`

        :param runtime: the pea to be attached
        :param args: not used (kept to maintain interface)
        :param kwargs: not used (kept to maintain interface)
        """
        self.runtime = runtime
        self.attached = True

    @property
    def req(self) -> 'Request':
        """Get the current (typed) request, shortcut to ``self.runtime.request``


        .. # noqa: DAR201
        """
        return self.runtime.request

    @property
    def partial_reqs(self) -> Sequence['Request']:
        """The collected partial requests under the current ``request_id``


        .. # noqa: DAR401


        .. # noqa: DAR201
        """
        if self.expect_parts > 1:
            return self.runtime.partial_requests
        else:
            raise ValueError(
                f'trying to access all partial requests, '
                f'but {self.runtime} has only one message'
            )

    @property
    def expect_parts(self) -> int:
        """The expected number of partial messages


        .. # noqa: DAR201
        """
        return self.runtime.expect_parts

    @property
    def docs(self) -> 'DocumentSet':
        """The DocumentSet after applying the traversal


        .. # noqa: DAR201"""
        from ..types.sets import DocumentSet

        if self.expect_parts > 1:
            return DocumentSet([d for r in reversed(self.partial_reqs) for d in r.docs])
        else:
            return self.req.docs

    @property
    def msg(self) -> 'Message':
        """Get the current request, shortcut to ``self.runtime.message``


        .. # noqa: DAR201
        """
        return self.runtime.message

    @property
    def queryset(self) -> 'QueryLangSet':
        """


        .. # noqa: DAR101


        .. # noqa: DAR102


        .. # noqa: DAR201
        """
        if self.msg:
            return self.msg.request.queryset
        else:
            return []

    @property
    def logger(self) -> 'JinaLogger':
        """Shortcut to ``self.runtime.logger``


        .. # noqa: DAR201
        """
        return self.runtime.logger

    def __call__(self, *args, **kwargs) -> None:
        """


        .. # noqa: DAR102


        .. # noqa: DAR101
        """
        raise NotImplementedError

    def __eq__(self, other):
        return self.__class__ == other.__class__

    def __getstate__(self) -> Dict[str, Any]:
        """
        Unlike `Executor`, driver is stateless.

        Therefore, on every save, it creates a new & empty driver object and save it.
        :return: the state in dict form
        """

        d = dict(self.__class__(**self._init_kwargs_dict).__dict__)
        return d


class ContextAwareRecursiveMixin:
    """
    The full data structure version of :class:`FlatRecursiveMixin`, to be mixed in with :class:`BaseRecursiveDriver`.
    It uses :meth:`traverse` in :class:`DocumentSet` and allows direct manipulation of Chunk-/Match-/DocumentSets.

    .. seealso::
       https://github.com/jina-ai/jina/issues/1932

    """

    def __call__(self, *args, **kwargs):
        """Traverse with _apply_all

        :param args: args forwarded to ``_apply_all``
        :param kwargs: kwargs forwarded to ``_apply_all``
        """
        document_sets = self.docs.traverse(self._traversal_paths)
        self._apply_all(document_sets, *args, **kwargs)

    def _apply_all(
        self,
        doc_sequences: Iterable['DocumentSet'],
        *args,
        **kwargs,
    ) -> None:
        """Apply function works on an Iterable of DocumentSet, modify the docs in-place.

        Each DocumentSet refers to a leaf (e.g. roots, matches or chunks wrapped
        in a :class:`jina.DocumentSet`) in the traversal_paths. Modifications on the
        DocumentSets (e.g. adding or deleting Documents) are directly applied on the underlying objects.
        Adding a chunk to a ChunkSet results in adding a chunk to the parent Document.

        :param doc_sequences: the Documents that should be handled
        :param args: driver specific arguments, which might be forwarded to the Executor
        :param kwargs: driver specific arguments, which might be forwarded to the Executor
        """


class FlatRecursiveMixin:
    """
    The batch optimized version of :class:`ContextAwareRecursiveMixin`, to be mixed in with :class:`BaseRecursiveDriver`.
    It uses :meth:`traverse_flattened_per_path` in :class:`DocumentSet` and yield much better performance
    when no context is needed and batching is possible.

    .. seealso::
       https://github.com/jina-ai/jina/issues/1932

    """

    def __call__(self, *args, **kwargs):
        """Traverse with _apply_all

        :param args: args forwarded to ``_apply_all``
        :param kwargs: kwargs forwarded to ``_apply_all``
        """
        path_documents = self.docs.traverse_flattened_per_path(self._traversal_paths)
        for documents in path_documents:
            if documents:
                self._apply_all(documents, *args, **kwargs)

    def _apply_all(
        self,
        docs: 'DocumentSet',
        *args,
        **kwargs,
    ) -> None:
        """Apply function works on a list of docs, modify the docs in-place.

        The list refers to all reachable leaves of a single ``traversal_path``.

        :param docs: the Documents that should be handled
        :param args: driver specific arguments, which might be forwarded to the Executor
        :param kwargs: driver specific arguments, which might be forwarded to the Executor

        """


class DocsExtractUpdateMixin:
    """
    A Driver pattern for extracting attributes from Documents, feeding to an executor and updating the Documents with
    the results.

    Drivers equipped with this mixin will have :method:`_apply_all` inherited.

    The :method:`_apply_all` implements the following logics:
        - From ``docs``, it extracts the attributes defined :method:`exec_fn`'s arguments.
        - It feeds the attributes to the bind executor's :method:`exec_fn`.
        - It updates ``docs`` with results returned from :method:`exec_fn`

    The following shortcut logics are implemented:
        - while extracting: attributes defined :method:`exec_fn`'s arguments are extracted from ``docs``;
        - while extracting: attributes annotated with ``ndarray`` are stacked into Numpy NdArray objects;
        - while updating: if ``exec_fn`` returns a List of Dict, then ``doc.set_attrs(**exec_result)`` is called;
        - while updating: if ``exec_fn`` returns a Document, then ``doc.update(exec_result)` is called.
        - while updating: if none of above applies, then calling :meth:`update_single_doc`

    To override the update behavior, you can choose to override:
        - :meth:`update_docs` if you want to modify the behavior of updating docs in bulk
        - :meth:`update_single_doc` if you want to modify the behavior of updating a single doc
    """

    def _apply_all(self, docs: 'DocumentSet') -> None:
        """Apply function works on a list of docs, modify the docs in-place.

        The list refers to all reachable leaves of a single ``traversal_path``.

        :param docs: the Documents that should be handled
        """

        contents, docs_pts = docs.extract_docs(
            *self._exec_fn_required_keys,
            stack_contents=self._exec_fn_required_keys_is_ndarray,
        )

        if docs_pts:
            if len(self._exec_fn_required_keys) > 1:
                exec_results = self.exec_fn(*contents)
            else:
                exec_results = self.exec_fn(contents)

            if exec_results is not None:
                # if exec_fn returns None then exec_fn is assumed to be immutable wrt. doc, hence skipped

                try:
                    len_results = len(exec_results)
                except:
                    try:
                        len_results = exec_results.shape[0]
                    except:
                        len_results = None

                if len(docs_pts) != len_results:
                    msg = (
                        f'mismatched {len(docs_pts)} docs from level {docs_pts[0].granularity} '
                        f'and length of returned: {len_results}, their length must be the same'
                    )
                    raise LengthMismatchException(msg)

                self.update_docs(docs_pts, exec_results)

    def update_docs(
        self,
        docs_pts: 'DocumentSet',
        exec_results: Union[List[Dict], List['Document'], Any],
    ) -> None:
        """
        Update Documents with the Executor returned results.

        :param: docs_pts: the set of document to be updated
        :param: exec_results: the results from :meth:`exec_fn`
        """
        from ..types.document import Document

        for doc, exec_result in zip(docs_pts, exec_results):
            if isinstance(exec_result, dict):
                doc.set_attrs(**exec_result)
            elif isinstance(exec_result, Document):
                # doc id should not be override with this method
                doc.update(exec_result, exclude_fields=('id',))
            else:
                self.update_single_doc(doc, exec_result)

    def update_single_doc(self, doc: 'Document', exec_result: Any) -> None:
        """Update a single Document with the Executor returned result.

        :param doc: the Document object
        :param exec_result: the single result from :meth:`exec_fn`
        """
        raise NotImplementedError

    @cached_property
    def _exec_fn_required_keys(self) -> List[str]:
        """Get the arguments of :attr:`exec_fn`.

        If ``strict_method_args`` set, then all arguments of :attr:`exec_fn` must be valid :class:`Document` attribute.

        :return: a list of supported arguments
        """

        if not self.exec_fn:
            raise ValueError(
                f'`exec_fn` is None, maybe {self} is not attached? call `self.attach`.'
            )

        required_keys = [
            k
            for k in inspect.getfullargspec(inspect.unwrap(self.exec_fn)).args
            if k != 'self'
        ]
        if not required_keys:
            raise AttributeError(f'{self.exec_fn} takes no argument.')

        if self._strict_method_args:
            from ..proto import jina_pb2
            from .. import Document

            support_keys = Document.get_all_attributes()
            unrecognized_keys = set(required_keys).difference(support_keys)
            if unrecognized_keys:
                camel_keys = set(
                    jina_pb2.DocumentProto().DESCRIPTOR.fields_by_camelcase_name
                )
                legacy_keys = {'data'}
                unrecognized_camel_keys = unrecognized_keys.intersection(camel_keys)
                if unrecognized_camel_keys:
                    raise AttributeError(
                        f'{unrecognized_camel_keys} are supported but you give them in CamelCase, '
                        f'please rewrite them in canonical form.'
                    )
                elif unrecognized_keys.intersection(legacy_keys):
                    raise AttributeError(
                        f'{unrecognized_keys.intersection(legacy_keys)} is now deprecated and not a valid argument of '
                        'the executor function, '
                        'please change `data` to `content: \'np.ndarray\'` in your executor function. '
                        'details: https://github.com/jina-ai/jina/pull/2313/'
                    )
                else:
                    raise AttributeError(
                        f'{unrecognized_keys} are invalid Document attributes, must come from {support_keys}'
                    )

        return required_keys

    @cached_property
    def _exec_fn_required_keys_is_ndarray(self) -> List[bool]:
        """Return a list of boolean indicators for showing if a key is annotated as ndarray

        :return: a list of boolean idicator, True if the corresponding key is annotated as ndarray
        """

        anno = inspect.getfullargspec(inspect.unwrap(self.exec_fn)).annotations
        return ['ndarray' in str(anno.get(k, '')) for k in self._exec_fn_required_keys]


class BaseRecursiveDriver(BaseDriver):
    """A :class:`BaseRecursiveDriver` is an abstract Driver class containing information about the `traversal_paths`
    that a `Driver` must apply its logic.
    It is intended to be mixed in with either :class:`FlatRecursiveMixin` or :class:`ContextAwareRecursiveMixin`
    """

    def __init__(self, traversal_paths: Tuple[str] = ('c', 'r'), *args, **kwargs):
        """Initialize a :class:`BaseRecursiveDriver`

        :param traversal_paths: Describes the leaves of the document tree on which _apply_all are called
        :param args: additional positional arguments which are just used for the parent initialization
        :param kwargs: additional key value arguments which are just used for the parent initialization
        """
        super().__init__(*args, **kwargs)
        self._traversal_paths = [path.lower() for path in traversal_paths]


class BaseExecutableDriver(BaseRecursiveDriver):
    """A :class:`BaseExecutableDriver` is an intermediate logic unit between the :class:`jina.peapods.runtimes.zmq.zed.ZEDRuntime` and :class:`jina.executors.BaseExecutor`
    It reads the protobuf message, extracts/modifies the required information and then sends to the :class:`jina.executors.BaseExecutor`,
    finally it returns the message back to :class:`jina.peapods.runtimes.zmq.zed.ZEDRuntime`.

    A :class:`BaseExecutableDriver` needs to be :attr:`attached` to a :class:`jina.peapods.runtimes.zmq.zed.ZEDRuntime` and :class:`jina.executors.BaseExecutor` before using.
    This is done by :func:`attach`. Note that a deserialized :class:`BaseDriver` from file is always unattached.
    """

    def __init__(
        self,
        executor: Optional[str] = None,
        method: Optional[str] = None,
        strict_method_args: bool = True,
        *args,
        **kwargs,
    ):
        """Initialize a :class:`BaseExecutableDriver`

        :param executor: the name of the sub-executor, only necessary when :class:`jina.executors.compound.CompoundExecutor` is used
        :param method: the function name of the executor that the driver feeds to
        :param strict_method_args: if set, then the input args of ``executor.method`` must be valid :class:`Document` attributes
        :param args: additional positional arguments which are just used for the parent initialization
        :param kwargs: additional key value arguments which are just used for the parent initialization
        """
        super().__init__(*args, **kwargs)
        self._executor_name = executor
        self._method_name = method
        self._strict_method_args = strict_method_args
        self._exec = None
        self._exec_fn = None

    @property
    def exec(self) -> 'AnyExecutor':
        """the executor that to which the instance is attached


        .. # noqa: DAR201
        """
        return self._exec

    @property
    def exec_fn(self) -> Callable:
        """the function of :func:`jina.executors.BaseExecutor` to call

        :return: the Callable to execute in the driver
        """
        if not self.runtime:
            return self._exec_fn
        elif (
            not self.msg.is_error
            or self.runtime.args.on_error_strategy < OnErrorStrategy.SKIP_EXECUTOR
        ):
            return self._exec_fn
        else:
            return lambda *args, **kwargs: None

    def attach(
        self, executor: 'AnyExecutor', req_type: Optional[str] = None, *args, **kwargs
    ) -> None:
        """Attach the driver to a :class:`jina.executors.BaseExecutor`

        :param executor: the executor to which we attach
        :param req_type: the request type to attach to
        :param args: additional positional arguments for the call of super().attach()
        :param kwargs: additional key value arguments for the call of super().attach()
        """
        super().attach(*args, **kwargs)
        if self._executor_name and isinstance(executor, CompoundExecutor):
            if self._executor_name in executor:
                self._exec = executor[self._executor_name]
            else:
                for c in executor.components:
                    if any(
                        t.__name__ == self._executor_name for t in type.mro(c.__class__)
                    ):
                        self._exec = c
                        break
            if self._exec is None:
                self.logger.critical(
                    f'fail to attach the driver to {executor}, '
                    f'no executor is named or typed as {self._executor_name}'
                )
        else:
            self._exec = executor

        if not self._method_name:
            decor_bindings = find_request_binding(self.exec.__class__)
            if req_type:
                canonic_name = _canonical_request_name(req_type)
            if req_type and canonic_name in decor_bindings:
                self._method_name = decor_bindings[canonic_name]
            elif 'default' in decor_bindings:
                self._method_name = decor_bindings['default']

        if self._method_name:
            self._exec_fn = getattr(self.exec, self._method_name)

    def __getstate__(self) -> Dict[str, Any]:
        """Do not save the executor and executor function, as it would be cross-referencing and unserializable.
        In other words, a deserialized :class:`BaseExecutableDriver` from file is always unattached.

        :return: dictionary of state
        """
        d = super().__getstate__()
        if '_exec' in d:
            del d['_exec']
        if '_exec_fn' in d:
            del d['_exec_fn']
        return d



import numpy as np
from jina.executors.classifiers import BaseClassifier
from jina.executors.decorators import batching, as_ndarray
from jina.executors.devices import TorchDevice


class TransformersTorchSeqClassifier(TorchDevice, BaseClassifier):
    """
    :class:`TransformersTorchSeqClassifier`
    Wrapper for :class:`transformers.AutoModelForSequenceClassification`. Works with all trained sequence classification models
    on `Huggingface Transformers model hub <https://huggingface.co/models?pipeline_tag=text-classification>`_.

    :param pretrained_model_name_or_path: Default: ``'distilbert-base-uncased-finetuned-sst-2-english'``. Can be either:

        - a string, the `model id` of a pretrained model hosted
            inside a model repo on huggingface.co, e.g.: ``bert-base-uncased``.
        - a path to a `directory` containing model weights saved using
            :func:`~transformers.PreTrainedModel.save_pretrained`, e.g.:
            ``./my_model_directory/``.

    :param args:  Additional positional arguments
    :param kwargs: Additional keyword arguments
    """

    def __init__(
        self,
        pretrained_model_name_or_path: str = 'distilbert-base-uncased-finetuned-sst-2-english',
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.pretrained_model_name_or_path = pretrained_model_name_or_path
        # self.max_length = max_length

    def post_init(self):
        """Load the transformer model and tokenizer"""
        from transformers import AutoTokenizer, AutoModelForSequenceClassification

        self.tokenizer = AutoTokenizer.from_pretrained(
            self.pretrained_model_name_or_path
        )
        self.model = AutoModelForSequenceClassification.from_pretrained(
            self.pretrained_model_name_or_path
        )
        self.to_device(self.model)

    @batching
    @as_ndarray
    def predict(self, content: 'np.ndarray', *args, **kwargs) -> 'np.ndarray':
        """
        Performs prediction with sequence classification model
        The size and type of output can be one of the follows (depending on the model used in initializing the class) ``B`` is ``content.shape[0]``:

            - (B,) or (B, 1); float
            - (B, L): soft label for L-class multi-class classification

        :param content: the input textual data to be classified, a 1 d
            array of string type in size `B`
        :type content: np.ndarray
        :param args:  Additional positional arguments
        :param kwargs: Additional keyword arguments
        :return: logits
        :rtype: np.ndarray
        """
        import torch

        inputs = self.tokenizer(
            list(content),
            add_special_tokens=True,
            return_tensors='pt',
            padding='longest',
            truncation=True,
        )
        with torch.no_grad():
            outputs = self.model(**inputs)[0]

        if self.model.config.num_labels == 1:
            scores = 1.0 / (1.0 + np.exp(-outputs))
        else:
            scores = np.exp(outputs) / np.exp(outputs).sum(-1, keepdims=True)

        scores = scores.detach().numpy()
        return scores



__copyright__ = "Copyright (c) 2020 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

from typing import Sequence, Any

from .. import BaseEvaluator


class BaseRankingEvaluator(BaseEvaluator):
    """A :class:`BaseRankingEvaluator` evaluates the content of matches against the expected GroundTruth.
    It is used to evaluate messages coming out from Indexers and Rankers and compares matches with groundtruths
    """

    def evaluate(
        self, actual: Sequence[Any], desired: Sequence[Any], *args, **kwargs
    ) -> float:
        """ "
        :param actual: the matched document identifiers from the request as matched by jina indexers and rankers
        :param desired: the expected documents matches ids sorted as they are expected
        :return the evaluation metric value for the request document
        """
        raise NotImplementedError



__copyright__ = "Copyright (c) 2021 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

from typing import Tuple

import numpy as np
from jina.executors.indexers.vector import BaseNumpyIndexer


class NGTIndexer(BaseNumpyIndexer):
    """
    NGT powered vector indexer

    For more information about the NGT supported parameters and installation problems, please consult:
        - https://github.com/yahoojapan/NGT

    .. note::
        NGT package dependency is only required at the query time.
        Quick Install : pip install ngt
    """

    def __init__(self, num_threads: int = 2, metric: str = 'L2', epsilon: float = 0.1, *args, **kwargs):
        """
        Initialize an NGT Indexer.

        :param num_threads: Number of threads to build index
        :param metric: Should be one of {L1,L2,Hamming,Jaccard,Angle,Normalized Angle,Cosine,Normalized Cosine}
        :param epsilon: Toggle this variable for speed vs recall tradeoff.
                        Higher value of epsilon means higher recall
                        but query time will increase with epsilon
        """

        super().__init__(*args, **kwargs)
        self._metric = metric
        self._num_threads = num_threads
        self._epsilon = epsilon

    def post_init(self):
        """Setup workspace."""

        super().post_init()
        self.index_path = self.get_file_from_workspace('index')

    def build_advanced_index(self, vecs: 'np.ndarray'):
        """
        Build an advanced index structure from a numpy array.

        :param vecs: numpy array containing the vectors to index
        :return: advanced NGT index
        """

        import ngtpy
        ngtpy.create(path=self.index_path, dimension=self.num_dim, distance_type=self._metric)
        _index = ngtpy.Index(self.index_path)
        _index.batch_insert(vecs, num_threads=self._num_threads)
        return _index

    def query(self, keys: 'np.ndarray', top_k: int, *args, **kwargs) -> Tuple['np.ndarray', 'np.ndarray']:
        """
        Find the top-k vectors with smallest ``metric`` and return their ids in ascending order.

        :param keys: numpy array containing vectors to search for
        :param top_k: upper limit of responses for each search vector
        :return: best matching indices and distance values in two separate `np.array`
        """

        dist = []
        idx = []
        for key in keys:
            results = self.query_handler.search(key, size=top_k, epsilon=self._epsilon)
            if results:
                index_k, distance_k = zip(*results)
                idx.append(index_k)
                dist.append(distance_k)

        return self._int2ext_id[np.array(idx)], np.array(dist)



from typing import Optional

from jina.drivers import BaseExecutableDriver


class DumpDriver(BaseExecutableDriver):
    """A Driver that calls the dump method of the Executor

    :param executor: the executor to which we attach the driver
    :param args: passed to super().__init__
    :param kwargs: passed to super().__init__
    """

    def __init__(
        self,
        executor: Optional[str] = None,
        *args,
        **kwargs,
    ):
        super().__init__(executor, 'dump', *args, **kwargs)

    def __call__(self, *args, **kwargs):
        """Call the Dump method of the Indexer to which the Driver is attached

        :param args: passed to the exec_fn
        :param kwargs: passed to the exec_fn
        """
        self.exec_fn(self.req.path, self.req.shards, *args, **kwargs)



__copyright__ = "Copyright (c) 2020 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

from typing import Optional

import numpy as np

from .. import BaseNumericEncoder
from ...decorators import batching
from ....excepts import UndefinedModel


class TransformEncoder(BaseNumericEncoder):
    """
    :class:`TransformEncoder` encodes data from an ndarray in size `B x T` into an ndarray in size `B x D`

    :param model_path: path from where to pickle the sklearn model.
    :param args: Extra positional arguments to be set
    :param kwargs: Extra keyword arguments to be set
    """

    def __init__(self, model_path: Optional[str] = None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.model_path = model_path

    def post_init(self) -> None:
        """Load the model from path if :param:`model_path` is set."""
        import pickle

        self.model = None
        if self.model_path:
            with open(self.model_path, 'rb') as model_file:
                self.model = pickle.load(model_file)

    @batching
    def encode(self, content: 'np.ndarray', *args, **kwargs) -> 'np.ndarray':
        """
        :param content: a `B x T` numpy ``ndarray``, `B` is the size of the batch
        :return: a `B x D` numpy ``ndarray``
        :param args: Extra positional arguments to be set
        :param kwargs: Extra keyword arguments to be set
        """
        return self.model.transform(content)



__copyright__ = "Copyright (c) 2020 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

import os
import pickle
from typing import Sequence

import numpy as np

from jina.executors.devices import TorchDevice
from jina.excepts import PretrainedModelFileDoesNotExist
from jina.executors.decorators import batching, as_ndarray
from jina.executors.encoders.multimodal import BaseMultiModalEncoder


class TirgMultiModalEncoder(TorchDevice, BaseMultiModalEncoder):

    def __init__(self, model_path: str = 'checkpoint.pth',
                 texts_path: str = 'texts.pkl',
                 channel_axis: int = -1,
                 positional_modality: Sequence[str] = ['image', 'text'],
                 *args, **kwargs):
        """
        :param model_path: the path where the model is stored.
        """
        super().__init__(positional_modality=positional_modality, *args, **kwargs)
        self.model_path = model_path
        self.texts_path = texts_path
        self.channel_axis = channel_axis
        # axis 0 is the batch
        self._default_channel_axis = 1

    def post_init(self):
        super().post_init()
        import torch
        from .img_text_composition_models import TIRG
        if self.model_path and os.path.exists(self.model_path):
            with open(self.texts_path, 'rb') as fp:
                texts = pickle.load(fp)
            self.model = TIRG(texts, 512)
            model_sd = torch.load(self.model_path, map_location=torch.device('cpu'))
            self.model.load_state_dict(model_sd['model_state_dict'])
            self.model.eval()
            self.to_device(self.model)
        else:
            raise PretrainedModelFileDoesNotExist(f'model {self.model_path} does not exist')

    def _get_features(self, content):
        import torch
        visual_data = content[(self.positional_modality.index('image'))]
        if self.channel_axis != self._default_channel_axis:
            visual_data = np.moveaxis(visual_data, self.channel_axis, self._default_channel_axis)
        textual_data = content[(self.positional_modality.index('text'))]

        visual_data = torch.from_numpy(np.stack(visual_data)).float()
        textual_data = np.stack(textual_data).tolist()

        if self.on_gpu:
            visual_data = visual_data.cuda()
            textual_data = textual_data.cuda()

        img_features = self.model.extract_img_feature(visual_data)
        text_features = self.model.extract_text_feature(textual_data)

        return self.model.compose_img_text_features(img_features, text_features)

    @batching(slice_nargs=2)
    @as_ndarray
    def encode(self, *content: 'np.ndarray', **kwargs) -> 'np.ndarray':
        feature = self._get_features(content).detach()
        if self.on_gpu:
            feature = feature.cpu()
        feature = feature.numpy()
        return feature



__copyright__ = "Copyright (c) 2021 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

from typing import List, Dict

import numpy as np
import torch
from jina.executors.decorators import single

from jina.executors.devices import TorchDevice
from jina.executors.segmenters import BaseSegmenter


class FaceNetSegmenter(TorchDevice, BaseSegmenter):
    """FaceNetSegmenter segments faces from an image.

    - Input shape: `(Height x Width x Channels)`
    - Output shape: `NumFaces x (Channels x ImageSize x ImageSize)`

    `Channels` dimension can be changed (e.g. set `channel_axis` to 0 for channels first mode instead of channels last).

    :param image_size: Height and width of a detected face. Smaller faces are upscaled.
    :param margin: Margin to add to bounding box, in terms of pixels in the final image.
    :param selection_method: Heuristic to use to select a single face from the image. Options:
        "probability": highest probability selected
        "largest": largest box selected
        "largest_over_threshold": largest box over a certain probability selected
        "center_weighted_size": box size minus weighted squared offset from image center
    :param post_process: Flag for normalizing the output image. Required if you want to pass
        these face to the FaceNetEmbedder.
    :param keep_all: Flag for keeping all detected faces rather than a single face.
    :param min_face_size: Minimum face size to search for.
    :param channel_axis: Axis of channels in the image. Default is 2 (channels-last), use 0 for channels-first.
    """

    def __init__(self,
                 image_size: int = 160,
                 margin: int = 0,
                 selection_method: str = 'largest',
                 post_process: bool = True,
                 keep_all: bool = False,
                 min_face_size: int = 20,
                 channel_axis: int = 2,
                 *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.image_size = image_size
        self.margin = margin
        self.selection_method = selection_method
        self.post_process = post_process
        self.keep_all = keep_all
        self.min_face_size = min_face_size
        self.channel_axis = channel_axis

        self._default_channel_axis = 2

    def post_init(self):
        from facenet_pytorch import MTCNN

        self.face_detector = MTCNN(selection_method=self.selection_method,
                                   image_size=self.image_size,
                                   margin=self.margin,
                                   device=self.device,
                                   post_process=self.post_process,
                                   min_face_size=self.min_face_size,
                                   keep_all=self.keep_all).eval()

    @single
    def segment(self, data: 'np.ndarray', *args, **kwargs) -> List[Dict]:
        """Transform a numpy `ndarray` of shape `(Height x Width x Channel)`
        into a list with dicts that contain cropped images.

        :param data: A numpy `ndarray` that represents a single image.
        :param args: Additional positional arguments.
        :param kwargs: Additional positional arguments.
        :return: A list with dicts that contain cropped images.
        """
        if self.channel_axis != self._default_channel_axis:
            data = np.moveaxis(data, self.channel_axis, self._default_channel_axis)

        with torch.no_grad():
            image = torch.from_numpy(data.astype('float32')).to(self.device)
            # Create a batch of size 1
            image = image.unsqueeze(0)

            faces, probabilities = self.face_detector(image, return_prob=True)

            if self.keep_all:
                # All faces and probabilities are grouped in the first dimension of the first batch element
                faces = faces[0]
                probabilities = probabilities[0]

            results = [
                dict(
                    offset=0,
                    weight=probability,
                    blob=face.detach().numpy(),
                )
                for face, probability in zip(faces, probabilities)
                if face is not None
            ]
            return results



__copyright__ = "Copyright (c) 2020 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

from collections import defaultdict
from typing import Dict, List, Callable, Union, Optional

from . import BaseExecutor, AnyExecutor


class CompoundExecutor(BaseExecutor):
    """A :class:`CompoundExecutor` is a set of multiple executors.
    The most common usage is chaining a pipeline of executors, where the
    input of the current is the output of the former.

    A common use case of :class:`CompoundExecutor` is to glue multiple :class:`BaseExecutor` together, instead of breaking them into different Pods.

    :param routes: a map of function routes. The key is the function name, the value is a tuple of two pieces,
        where the first element is the name of the referred component (``metas.name``) and the second element
        is the name of the referred function.

        .. seealso::

            :func:`add_route`
    :param resolve_all: universally add ``*_all()`` to all functions that have the identical name

    **Example 1: a compound Chunk Indexer that does vector indexing and key-value index**

    .. highlight:: yaml
    .. code-block:: yaml

        !CompoundExecutor
        components:
          - !NumpyIndexer
            with:
              index_filename: vec.gz
            metas:
              name: vecidx_exec  # a customized name
              workspace: ${{TEST_WORKDIR}}
          - !BinaryPbIndexer
            with:
              index_filename: chunk.gz
            metas:
              name: chunkidx_exec
              workspace: ${{TEST_WORKDIR}}
        metas:
          name: chunk_compound_indexer
          workspace: ${{TEST_WORKDIR}}
        requests:
          on:
            SearchRequest:
              - !VectorSearchDriver
                with:
                  executor: vecidx_exec
            IndexRequest:
              - !VectorIndexDriver
                with:
                  executor: vecidx_exec
            ControlRequest:
              - !ControlReqDriver {}

    **Example 2: a compound crafter that first craft the doc and then segment **

    .. highlight:: yaml
    .. code-block:: yaml

        !CompoundExecutor
        components:
          - !GifNameRawSplit
            metas:
              name: name_split  # a customized name
              workspace: ${{TEST_WORKDIR}}
          - !GifPreprocessor
            with:
              every_k_frame: 2
              from_buffer: true
            metas:
              name: gif2chunk_preprocessor  # a customized name
        metas:
          name: compound_crafter
          workspace: ${{TEST_WORKDIR}}
          py_modules: gif2chunk.py
        requests:
          on:
            IndexRequest:
              - !DocCraftDriver
                with:
                  executor: name_split
              - !SegmentDriver
                with:
                  executor: gif2chunk_preprocessor
            ControlRequest:
              - !ControlReqDriver {}

    Create a new :class:`CompoundExecutor` object


     **Example 3: **

        We have two dummy executors as follows:

        .. highlight:: python
        .. code-block:: python

            class dummyA(BaseExecutor):
                def say(self):
                    return 'a'

                def sayA(self):
                    print('A: im A')


            class dummyB(BaseExecutor):
                def say(self):
                    return 'b'

                def sayB(self):
                    print('B: im B')

        and we create a :class:`CompoundExecutor` consisting of these two via

        .. highlight:: python
        .. code-block:: python

            da, db = dummyA(), dummyB()
            ce = CompoundExecutor()
            ce.components = lambda: [da, db]

        Now the new executor ``ce`` have two new methods, i.e :func:`ce.sayA` and :func:`ce.sayB`. They point to the original
        :func:`dummyA.sayA` and :func:`dummyB.sayB` respectively. One can say ``ce`` has inherited these two methods.

        The interesting part is :func:`say`, as this function name is shared between :class:`dummyA` and :class:`dummyB`.
        It requires some resolution. When `resolve_all=True`, then a new function :func:`say_all` is add to ``ce``.
        ``ce.say_all`` works as if you call :func:`dummyA.sayA` and :func:`dummyB.sayB` in a row. This
        makes sense in some cases such as training, saving. In other cases, it may require a more sophisticated resolution,
        where one can use :func:`add_route` to achieve that. For example,

        .. highlight:: python
        .. code-block:: python

            ce.add_route('say', db.name, 'say')
            assert b.say() == 'b'

        Such resolution is what we call **routes** here, and it can be specified in advance with the
        arguments ``routes`` in :func:`__init__`, or using YAML.

        .. highlight:: yaml
        .. code-block:: yaml

            !CompoundExecutor
            components: ...
            with:
              resolve_all: true
              routes:
                say:
                - dummyB-e3acc910
                - say

        .. warning::

            When setting inner `executors` in `components` the `workspace` configuration will not be used and will be overridden
            by a workspace extracted considering the name of the `CompoundExecutor`, the name of each internal `Component` and the `pea_id`


    One can access the component of a :class:`CompoundExecutor` via index, e.g.

    .. highlight:: python
    .. code-block:: python

        c = BaseExecutor.load_config('compound-example.yaml')
        assertTrue(c[0] == c['dummyA-1ef90ea8'])
        c[0].add(obj)

    .. note::
        Component ``workspace`` and ``pea_id`` are overridden by their :class:`CompoundExecutor` counterparts.

    .. warning::

        When sub-component is external, ``py_modules`` must be given at root level ``metas`` not at the sub-level.

    """

    class _FnWrapper:
        def __init__(self, fns):
            self.fns = fns

        def __call__(self, *args, **kwargs):
            r = []
            for f in self.fns:
                r.append(f())
            return r

    class _FnAllWrapper(_FnWrapper):
        def __call__(self, *args, **kwargs):
            return all(super().__call__(*args, **kwargs))

    class _FnOrWrapper(_FnWrapper):
        def __call__(self, *args, **kwargs):
            return any(super().__call__(*args, **kwargs))

    def __init__(
        self, routes: Dict[str, Dict] = None, resolve_all: bool = True, *args, **kwargs
    ):
        super().__init__(*args, **kwargs)
        self._components = None  # type: Optional[List[AnyExecutor]]
        self._routes = routes
        self._is_updated = False  #: the internal update state of this compound executor
        self.resolve_all = resolve_all

    @property
    def is_updated(self) -> bool:
        """
        Return ``True``  if any components is updated.

        :return: only true if all components are updated or if the compound is updated
        """
        return (
            self.components and any(c.is_updated for c in self.components)
        ) or self._is_updated

    @is_updated.setter
    def is_updated(self, val: bool) -> None:
        """
        Set :attr:`is_updated` for this :class:`CompoundExecutor`. Note, not to all its components

        :param val: new value of :attr:`is_updated`
        """
        self._is_updated = val

    def save(self, filename: Optional[str] = None):
        """
        Serialize this compound executor along with all components in it to binary files.
        It uses ``pickle`` for dumping.

        :param filename: file path of the serialized file, if not given then :attr:`save_abspath` is used
        """
        for c in self.components:
            c.save()
        super().save(
            filename=filename
        )  # do i really need to save the compound executor itself

    @property
    def components(self) -> List[AnyExecutor]:
        """
        Return all component executors as a list. The list follows the order as defined in the YAML config or the
        pre-given order when calling the setter.

        :return: components
        """
        return self._components

    @components.setter
    def components(self, comps: Callable[[], List]) -> None:
        """Set the components of this executors

        :param comps: a function returns a list of executors
        """
        if not callable(comps):
            raise TypeError(
                'components must be a callable function that returns '
                'a List[BaseExecutor]'
            )

        # Important to handle when loading a CompoundExecutor when `inner` executors have not been loaded from yaml
        if not getattr(self, '_init_from_yaml', False):
            self._components = comps()
            if not isinstance(self._components, list):
                raise TypeError(
                    f'components expect a list of executors, receiving {type(self._components)!r}'
                )
            self._set_comp_workspace()
            self._resolve_routes()
            self._post_components()
        else:
            self.logger.debug(
                'components is omitted from construction, as it is initialized from yaml config'
            )

    @staticmethod
    def get_component_workspace_from_compound_workspace(
        compound_workspace: str, compound_name: str, pea_id: int
    ) -> str:
        """
        Get the name of workspace.

        :param compound_workspace: Workspace of the compound executor.
        :param compound_name: Name of the compound executor.
        :param pea_id: Id of the pea.
        :return: The name of workspace.
        """
        import os

        return (
            BaseExecutor.get_shard_workspace(compound_workspace, compound_name, pea_id)
            if (isinstance(pea_id, int) and pea_id > 0)
            else os.path.join(compound_workspace, compound_name)
        )

    def _set_comp_workspace(self) -> None:
        # overrides the workspace setting for all components
        for c in self.components:
            if not c.workspace and self.workspace:
                c_workspace = (
                    CompoundExecutor.get_component_workspace_from_compound_workspace(
                        self.workspace, self.name, self.pea_id
                    )
                )
                self.logger.warning(f'Setting workspace of {c.name} to {c_workspace}')
                c.workspace = c_workspace

    def _resolve_routes(self) -> None:
        if self._routes:
            for f, v in self._routes.items():
                for kk, vv in v.items():
                    self.add_route(f, kk, vv)

    def add_route(
        self, fn_name: str, comp_name: str, comp_fn_name: str, is_stored: bool = False
    ) -> None:
        """Create a new function for this executor which refers to the component's function

        This will create a new function :func:`fn_name` which actually refers to ``components[comp_name].comp_fn_name``.
        It is useful when two components have a function with duplicated name and one wants to resolve this duplication.

        :param fn_name: the name of the new function
        :param comp_name: the name of the referred component, defined in ``metas.name``
        :param comp_fn_name: the name of the referred function of ``comp_name``
        :param is_stored: if ``True`` then this change will be stored in the config and affects future :func:`save` and
            :func:`save_config`

        """
        for c in self.components:
            if (
                c.name == comp_name
                and hasattr(c, comp_fn_name)
                and callable(getattr(c, comp_fn_name))
            ):
                setattr(self, fn_name, getattr(c, comp_fn_name))
                if is_stored:
                    if not self._routes:
                        self._routes = {}
                    self._routes[fn_name] = {comp_name: comp_fn_name}
                    self.is_updated = True
                return
        else:
            raise AttributeError(f'bad names: {comp_name} and {comp_fn_name}')

    def close(self) -> None:
        """Close all components and release the resources"""
        if self.components:
            for c in self.components:
                c.close()
        super().close()

    def __contains__(self, item: str):
        if isinstance(item, str):
            for c in self.components:
                if c.name == item:
                    return True
            return False
        else:
            raise TypeError('CompoundExecutor only support string type "in"')

    def __getitem__(self, item: Union[int, str]):
        if isinstance(item, int):
            return self.components[item]
        elif isinstance(item, str):
            for c in self.components:
                if c.name == item:
                    return c
        else:
            raise TypeError('CompoundExecutor only supports int or string index')

    def __iter__(self):
        return self.components.__iter__()

    def _post_components(self):
        pass



"""Module wrapping interactions with the remote Jina Hub API"""
from jina.logging.logger import JinaLogger
import json
import base64
from urllib.error import HTTPError
from urllib.parse import urlencode
from urllib.request import Request, urlopen
from pkg_resources import resource_stream
from typing import Dict, Sequence, Any, Optional, List, Tuple


from .local import (
    _fetch_access_token,
    _make_hub_table,
    _make_hub_table_with_local,
    _load_local_hub_manifest,
)
from ...jaml import JAML
from ...helper import colored
from ...importer import ImportExtensions
from ...logging.profile import TimeContext
from ...excepts import HubLoginRequired


def _list(
    logger: JinaLogger,
    image_name: Optional[str] = None,
    image_kind: Optional[str] = None,
    image_type: Optional[str] = None,
    image_keywords: Sequence = (),
) -> Optional[List[Dict[str, Any]]]:
    """Use Hub api to get the list of filtered images.

    :param logger: logger to use
    :param image_name: name of hub image
    :param image_kind: kind of hub image (indexer/encoder/segmenter/crafter/evaluator/ranker etc)
    :param image_type: type of hub image (pod/app)
    :param image_keywords: keywords added in the manifest yml
    :return: a dict of manifest specifications, each coresponds to a hub image
    """
    with resource_stream('jina', '/'.join(('resources', 'hubapi.yml'))) as fp:
        hubapi_yml = JAML.load(fp)
        hubapi_url = hubapi_yml['hubapi']['url'] + hubapi_yml['hubapi']['list']

    params = {
        'name': image_name,
        'kind': image_kind,
        'type': image_type,
        'keywords': image_keywords,
    }
    params = {k: v for k, v in params.items() if v}
    if params:
        data = urlencode(params, doseq=True)
        request = Request(f'{hubapi_url}?{data}')
        with TimeContext('searching', logger):
            try:
                with urlopen(request) as resp:
                    response = json.load(resp)
            except HTTPError as err:
                if err.code == 400:
                    logger.warning(
                        'no matched executors found. please use different filters and retry.'
                    )
                elif err.code == 500:
                    logger.error(f'server is down: {err.reason}')
                else:
                    logger.error(f'unknown error: {err.reason}')
                return

        local_manifest = _load_local_hub_manifest()
        if local_manifest:
            tb = _make_hub_table_with_local(response, local_manifest)
        else:
            tb = _make_hub_table(response)
        logger.info('\n'.join(tb))
        return response


def _fetch_docker_auth(logger: JinaLogger) -> Tuple[str, str]:
    """Use Hub api to get docker credentials.

    :param logger: the logger instance
    :return: a dict of specifying username and password
    """
    with resource_stream('jina', '/'.join(('resources', 'hubapi.yml'))) as fp:
        hubapi_yml = JAML.load(fp)
        hubapi_url = hubapi_yml['hubapi']['url'] + hubapi_yml['hubapi']['docker_auth']

    with ImportExtensions(
        required=True,
        help_text='Missing "requests" dependency, please do pip install "jina[http]"',
    ):
        import requests

    headers = {
        'Accept': 'application/json',
        'authorizationToken': _fetch_access_token(logger),
    }
    response = requests.get(url=f'{hubapi_url}', headers=headers)
    if response.status_code != requests.codes.ok:
        raise HubLoginRequired(
            f'❌ Failed to fetch docker credentials. status code {response.status_code}'
        )
    json_response = json.loads(response.text)
    username = base64.b64decode(json_response['docker_username']).decode('ascii')
    password = base64.b64decode(json_response['docker_password']).decode('ascii')
    logger.debug(f'✅ Successfully fetched docker creds for user')
    return username, password


def _register_to_mongodb(logger: JinaLogger, summary: Optional[Dict] = None):
    """Hub API Invocation to run `hub push`.

    :param logger: the logger instance
    :param summary: the summary dict object
    """
    # TODO(Deepankar): implement to jsonschema based validation for summary
    logger.info('registering image to Jina Hub database...')

    with resource_stream('jina', '/'.join(('resources', 'hubapi.yml'))) as fp:
        hubapi_yml = JAML.load(fp)
        hubapi_url = hubapi_yml['hubapi']['url'] + hubapi_yml['hubapi']['push']

    with ImportExtensions(
        required=True,
        help_text='Missing "requests" dependency, please do pip install "jina[http]"',
    ):
        import requests

    headers = {
        'Accept': 'application/json',
        'authorizationToken': _fetch_access_token(logger),
    }
    response = requests.post(
        url=f'{hubapi_url}', headers=headers, data=json.dumps(summary)
    )
    if response.status_code == requests.codes.ok:
        logger.success(f'✅ Successfully updated the database. {response.text}')
    else:
        raise HubLoginRequired(
            f'❌ Got an error from the API: {response.text.rstrip()}. '
            f'Please login using command: {colored("jina hub login", attrs=["bold"])}'
        )



__copyright__ = "Copyright (c) 2021 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

from typing import Dict

import numpy as np

from jina.executors.decorators import single
from jina.executors.crafters import BaseCrafter

from .helper import _crop_image, _move_channel_axis, _load_image


class ImageCropper(BaseCrafter):
    """
    Crop the image with the specific crop box.

    The coordinate is the same coordinate-system in
        the :py:mode:`PIL.Image`.

    :param top: the vertical coordinate of the top left
        corner of the crop box.
    :param left: the horizontal coordinate of the top
        left corner of the crop box.
    :param height: the height of the crop box.
    :param width: the width of the crop box.
    :param channel_axis: the axis referring to the channels
    :param args:  Additional positional arguments
    :param kwargs: Additional keyword arguments
    """

    def __init__(self, top: int = 0, left: int = 0, height: int = 224, width: int = 224, channel_axis: int = -1, *args, **kwargs):
        """Set Constructor."""
        super().__init__(*args, **kwargs)
        self.top = top
        self.left = left
        self.height = height
        self.width = width
        self.channel_axis = channel_axis

    @single
    def craft(self, blob: 'np.ndarray', *args, **kwargs) -> Dict:
        """
        Crop the input image array.

        :param blob: The ndarray of the image
        :return: A dict with the cropped image
        """
        raw_img = _load_image(blob, self.channel_axis)
        _img, top, left = _crop_image(raw_img, target_size=(self.height, self.width), top=self.top, left=self.left)
        img = _move_channel_axis(np.asarray(_img), -1, self.channel_axis)
        return dict(offset=0, blob=img.astype('float32'), location=(top, left))




__copyright__ = "Copyright (c) 2020 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"
from jina.executors.encoders.numeric import TransformEncoder


class RandomGaussianEncoder(TransformEncoder):
    """
    Reduce dimensionality using Gaussian random projection.

    Encodes ``Document`` content from an ndarray in size `B x T` into an ndarray in size `B x D`
    Where `B` is the batch's size and `T` and `D` are the dimensions pre (`T`)
    and after (`D`) the encoding.

    More info can be found
    `here <https://scikit-learn.org/stable/modules/generated/sklearn.random_projection.GaussianRandomProjection.html>`_
    """



__copyright__ = "Copyright (c) 2020 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

from typing import Any

from .. import BaseExecutor

if False:
    # fix type-hint complain for sphinx and flake
    from typing import TypeVar
    import numpy as np
    import scipy
    import tensorflow as tf
    import torch

    EncodingType = TypeVar(
        'EncodingType',
        np.ndarray,
        scipy.sparse.csr_matrix,
        scipy.sparse.coo_matrix,
        scipy.sparse.bsr_matrix,
        scipy.sparse.csc_matrix,
        torch.sparse_coo_tensor,
        tf.SparseTensor,
    )


class BaseEncoder(BaseExecutor):
    """``BaseEncoder`` encodes chunk into vector representation.

    The key function is :func:`encode`.

    .. seealso::
        :mod:`jina.drivers.encode`
    """

    def encode(self, content: 'np.ndarray', *args, **kwargs) -> 'EncodingType':
        """Encode the data, needs to be implemented in subclass.
        :param content: the data to be encoded
        :param args: additional positional arguments
        :param kwargs: additional key-value arguments
        """

        raise NotImplementedError


class BaseNumericEncoder(BaseEncoder):
    """BaseNumericEncoder encodes data from a ndarray, potentially B x ([T] x D) into a ndarray of B x D"""

    def encode(self, content: 'np.ndarray', *args, **kwargs) -> 'EncodingType':
        """
        :param content: a `B x ([T] x D)` numpy ``ndarray``, `B` is the size of the batch
        :param args: additional positional arguments
        :param kwargs: additional key-value arguments
        """
        raise NotImplementedError


class BaseImageEncoder(BaseNumericEncoder):
    """BaseImageEncoder encodes data from a ndarray, potentially B x (Height x Width) into a ndarray of B x D"""

    pass


class BaseVideoEncoder(BaseNumericEncoder):
    """BaseVideoEncoder encodes data from a ndarray, potentially B x (Time x Height x Width) into a ndarray of B x D"""

    pass


class BaseAudioEncoder(BaseNumericEncoder):
    """BaseAudioEncoder encodes data from a ndarray, potentially B x (Time x D) into a ndarray of B x D"""

    pass


class BaseTextEncoder(BaseEncoder):
    """
    BaseTextEncoder encodes data from an array of string type (data.dtype.kind == 'U') of size B into a ndarray of B x D.
    """

    def encode(self, content: 'np.ndarray', *args, **kwargs) -> 'EncodingType':
        """

        :param content: an 1d array of string type (data.dtype.kind == 'U') in size B
        :param args: additional positional arguments
        :param kwargs: additional key-value arguments
        """
        raise NotImplementedError



from ..base import set_base_parser
from ..helper import _chf


def set_hub_pushpull_parser(parser=None):
    """Set the parser for the hub push or hub pull

    :param parser: an optional existing parser to build upon
    :return: the parser
    """
    if not parser:
        parser = set_base_parser()

    from .login import mixin_hub_docker_login_parser
    from .pushpull import mixin_hub_pushpull_parser

    mixin_hub_docker_login_parser(parser)
    mixin_hub_pushpull_parser(parser)
    return parser


def set_hub_build_parser(parser=None):
    """Set the parser for `hub build`

    :param parser: the parser configure
    :return: the new parser
    """
    if not parser:
        parser = set_base_parser()

    from .login import mixin_hub_docker_login_parser
    from .build import mixin_hub_build_parser

    mixin_hub_docker_login_parser(parser)
    mixin_hub_build_parser(parser)

    return parser


def set_hub_list_parser(parser=None):
    """Set the parser for `hub list`

    :param parser: the parser configure
    :return: the new parser
    """
    if not parser:
        parser = set_base_parser()

    from .list import mixin_hub_list_parser

    mixin_hub_list_parser(parser)

    return parser


def set_hub_new_parser(parser=None):
    """Set the parser for the `hub new` command

    :param parser: the parser configure
    :return: the new parser
    """
    if not parser:
        parser = set_base_parser()

    from .new import mixin_hub_new_parser

    mixin_hub_new_parser(parser)

    return parser


def set_hub_parser(parser=None):
    """Set the parser for the hub

    :param parser: the parser configure
    """
    if not parser:
        parser = set_base_parser()

    spp = parser.add_subparsers(
        dest='hub',
        description='use `%(prog)-8s [sub-command] --help` '
        'to get detailed information about each sub-command',
        required=True,
    )

    spp.add_parser(
        'login',
        help='login via Github to push images to Jina hub registry',
        description='Login via Github to push images to Jina hub registry',
        formatter_class=_chf,
    )

    set_hub_new_parser(
        spp.add_parser(
            'new',
            aliases=['init', 'create'],
            help='create a new Hub executor or app using cookiecutter',
            description='Create a new Hub executor or app using cookiecutter',
            formatter_class=_chf,
        )
    )

    set_hub_build_parser(
        spp.add_parser(
            'build',
            help='build a directory into Jina hub image',
            description='Build a directory into Jina hub image',
            formatter_class=_chf,
        )
    )

    set_hub_pushpull_parser(
        spp.add_parser(
            'push',
            help='push an image to the Jina hub registry',
            description='Push an image to the Jina hub registry',
            formatter_class=_chf,
        )
    )

    set_hub_pushpull_parser(
        spp.add_parser(
            'pull',
            help='pull an image from the Jina hub registry to local',
            description='Pull an image to the Jina hub registry to local',
            formatter_class=_chf,
        )
    )

    set_hub_list_parser(
        spp.add_parser(
            'list',
            help='list hub executors from jina hub registry',
            description='List hub executors from jina hub registry',
            formatter_class=_chf,
        )
    )



from typing import Sequence
from .traversable import TraversableSequence

if False:
    from ..document.helper import DocGroundtruthPair


class DocumentGroundtruthSequence(TraversableSequence):
    def __init__(self, pairs: Sequence['DocGroundtruthPair']):
        self._pairs = pairs

    def __iter__(self):
        yield from self._pairs



__copyright__ = "Copyright (c) 2020 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

from typing import Sequence

import numpy as np

from ... import BaseExecutor


class BaseMultiModalEncoder(BaseExecutor):
    """
    :class:`BaseMultiModalEncoder` encodes data from multiple inputs (``text``, ``buffer``, ``blob`` or other ``embeddings``)
    into a single ``embedding``
    """

    def __init__(self, positional_modality: Sequence[str], *args, **kwargs):
        """
        :param positional_modality: the list of arguments indicating in which order the modalities they need to come
        for the encoding method
        :return:
        """
        super().__init__(*args, **kwargs)
        self.positional_modality = positional_modality

    def encode(self, *data: 'np.ndarray', **kwargs) -> 'np.ndarray':
        """
        :param: data: M arguments of shape `B x (D)` numpy ``ndarray``, `B` is the size of the batch, `M` is the number of modalities
        :return: a `B x D` numpy ``ndarray``
        """
        raise NotImplementedError



__copyright__ = "Copyright (c) 2020 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

from typing import Optional, Tuple, Any, Dict, List

from . import BaseExecutableDriver, FlatRecursiveMixin, DocsExtractUpdateMixin
from ..types.document import Document


class SegmentDriver(DocsExtractUpdateMixin, FlatRecursiveMixin, BaseExecutableDriver):
    """Drivers inherited from this Driver will bind :meth:`segment` by default """

    def __init__(
        self,
        executor: Optional[str] = None,
        method: str = 'segment',
        traversal_paths: Tuple[str] = ('r',),
        *args,
        **kwargs,
    ):
        super().__init__(
            executor, method, traversal_paths=traversal_paths, *args, **kwargs
        )

    def update_single_doc(self, doc: 'Document', exec_result: List[Dict]) -> None:
        """Update the document's chunks field with executor's returns.

        :param doc: the Document object
        :param exec_result: the single result from :meth:`exec_fn`
        """
        new_chunks = []
        for chunk in exec_result:
            with Document(**chunk) as c:
                if not c.mime_type:
                    c.mime_type = doc.mime_type
            new_chunks.append(c)
        doc.chunks.extend(new_chunks)



__copyright__ = "Copyright (c) 2021 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

from typing import Optional, Iterable

from jina import Document
from jina.executors.indexers.keyvalue import BinaryPbIndexer

if False:
    from plyvel import DB


class LevelDBIndexer(BinaryPbIndexer):
    """
    :class:`LevelDBIndexer` use `LevelDB` to save and query protobuf document.
    """

    def get_add_handler(self) -> 'DB':
        """Get the database handler."""
        import plyvel
        return plyvel.DB(self.index_abspath, create_if_missing=True)

    def get_create_handler(self) -> 'DB':
        """Get the database handler."""
        import plyvel
        return plyvel.DB(self.index_abspath, create_if_missing=True)

    def get_query_handler(self) -> 'DB':
        """Get the database handler."""
        import plyvel
        return plyvel.DB(self.index_abspath, create_if_missing=True)

    def query(self, key: str, *args, **kwargs) -> Optional[bytes]:
        """Find the serialized protobuf documents via id.

        :param key: ``id``
        :return: serialized document
        """
        from google.protobuf.json_format import Parse
        v = self.query_handler.get(bytes(key))
        value = None
        if v is not None:
            value = Parse(v.decode('utf8'), Document())
        return value

    def add(self, keys: Iterable[str], values: Iterable[bytes], *args, **kwargs) -> None:
        """Add JSON-friendly serialized documents to the index.

        :param keys: document ids
        :param values: serialized documents
        """
        with self.query_handler.write_batch() as h:
            for k, v in zip(keys, values):
                h.put(bytes(k), v)

    def update(self, keys: Iterable[str], values: Iterable[bytes], *args, **kwargs) -> None:
        """Update serialized documents on the index.

        :param keys: document ids to update
        :param values: serialized documents
        """
        missed = []
        for key in keys:
            if self.query_handler.get(bytes(key)) is None:
                missed.append(key)
        if missed:
            raise KeyError(f'Key(s) {missed} were not found in {self.save_abspath}')

        self.add(keys, values)

    def delete(self, keys: Iterable[str], *args, **kwargs) -> None:
        """Delete documents from the index.

        :param keys: document ids to delete
        """
        with self.query_handler.write_batch() as h:
            for k in keys:
                h.delete(bytes(k))



import numpy as np

from typing import List, Optional

from jina.executors.classifiers import BaseClassifier
from jina.executors.decorators import batching, as_ndarray
from jina.executors.devices import TFDevice


class ZeroShotTFClassifier(TFDevice, BaseClassifier):
    """
    :class: 'ZeroShotTFClassifier' wraps tensorflow transformers from
        huggingface, and performs zero shot learning classification
        for nlp data.

    :param labels: the potential labels for the classification
        task.
    :param pretrained_model_name_or_path: Either:
        - a string, the 'model id' of a pretrained model hosted inside a
            model repo on huggingface.co, e.g.: 'bert-base-uncased'.
        - a path to a 'directory' containing model weights saved using
            :func: '~transformers.PreTrainedModel.save_pretrained',
            e.g.: './my_model_directory/'.
    :param base_tokenizer_model: The name of the base model to use for
        creating the tokenizer. If None, will be equal to
        'pretrained_model_name_or_path'.
    :param pooling_strategy: the strategy to merge the word embeddings
        into the chunk embedding. Supported strategies include 'cls',
        'mean', 'max', 'min'.
    :param layer_index: index of the transformer layer that is used
        to create encodings. Layer 0 corresponds to the embeddings layer
    :param max_length: the max length to truncate the tokenized
        sequences to.
    """

    def __init__(
        self,
        labels: List[str],
        pretrained_model_name_or_path: str = 'distilbert-base-uncased',
        base_tokenizer_model: Optional[str] = None,
        pooling_strategy: str = 'mean',
        layer_index: int = -1,
        max_length: Optional[int] = None,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.labels = labels
        self.pretrained_model_name_or_path = pretrained_model_name_or_path
        self.base_tokenizer_model = (
            base_tokenizer_model or pretrained_model_name_or_path
        )
        self.pooling_strategy = pooling_strategy
        self.layer_index = layer_index
        self.max_length = max_length

        if self.pooling_strategy == 'auto':
            self.pooling_strategy = 'cls'
            self.logger.warning(
                '`auto` pooling_strategy is deprecated, Defaulting to '
                ' `cls` to maintain the old default behavior.'
            )

        if self.pooling_strategy not in ['cls', 'mean', 'max', 'min']:
            self.logger.error(
                f'pooling strategy not found: {self.pooling_strategy}.'
                ' The allowed pooling strategies are'
                ' `cls`, `mean`, `max`, `min`.'
            )
            raise NotImplementedError

        if len(self.labels) < 2:
            raise ValueError('The number of target labels must be at least 2.')

        if len(self.labels) != \
                len(set(self.labels)):
            raise ValueError(
                'There are duplicate value in the target_label argument.'
            )

    @as_ndarray
    @batching
    def predict(self,
                content: 'np.ndarray',
                *args,
                **kwargs) -> 'np.ndarray':
        """
         Perform zero shot classification on ``Document`` content, the predicted label
         for each sample in X is returned.

         The output is a zero/one one-hot label for L-class multi-class
         classification of size (B, L) with 'B' being 'content.shape[0]'
         and 'L' being the number of potential classification labels.

         :param content: the input textual data to be classified, a 1 d
            array of string type in size 'B'
         :type content: np.ndarray
         :return: zero/one one-hot predicted label of each sample
            in size '(B, L)'
         :rtype: np.ndarray
        """

        data_encoded = self._encode(content)

        distances = self._evaluate(data_encoded, self.labels_encoded)

        labels_pred = (distances == distances.min(axis=1)[:, None]).astype(int)

        return labels_pred

    def post_init(self):
        from transformers import TFAutoModel, AutoTokenizer

        self.tokenizer = AutoTokenizer.from_pretrained(
            self.base_tokenizer_model
        )
        self.model = TFAutoModel.from_pretrained(
            self.pretrained_model_name_or_path, output_hidden_states=True
        )
        self.to_device()

        self.labels_encoded = self._encode(self.labels)

    def _evaluate(self,
                  actual: 'np.ndarray',
                  desired: 'np.ndarray') -> 'np.ndarray':

        actual = _expand_vector(actual)
        desired = _expand_vector(desired)

        return _cosine(_ext_A(_norm(actual)), _ext_B(_norm(desired)))

    def _encode(self, content: 'np.ndarray', *args, **kwargs) -> 'np.ndarray':

        import tensorflow as tf

        if not self.tokenizer.pad_token:
            self.tokenizer.add_special_tokens({'pad_token': '[PAD]'})
            self.model.resize_token_embeddings(len(self.tokenizer.vocab))

        input_tokens = self.tokenizer(
            list(content),
            max_length=self.max_length,
            padding='longest',
            truncation=True,
            return_tensors='tf',
        )

        outputs = self.model(**input_tokens)

        n_layers = len(outputs.hidden_states)
        if self.layer_index not in list(range(-n_layers, n_layers)):
            self.logger.error(
                f'Invalid value {self.layer_index} for `layer_index`,'
                f' for the model {self.pretrained_model_name_or_path}'
                f' valid values are integers from {-n_layers} '
                f' to {n_layers - 1}.'
            )
            raise ValueError

        if self.pooling_strategy == 'cls' and not self.tokenizer.cls_token:
            self.logger.error(
                f'You have set pooling_strategy to `cls`, but the tokenizer'
                f' for the model {self.pretrained_model_name_or_path}'
                f' does not have a cls token set.'
            )
            raise ValueError

        fill_vals = {'cls': 0.0, 'mean': 0.0, 'max': -np.inf, 'min': np.inf}
        fill_val = tf.constant(fill_vals[self.pooling_strategy])

        layer = outputs.hidden_states[self.layer_index]
        attn_mask = tf.expand_dims(input_tokens['attention_mask'], -1)
        attn_mask = tf.broadcast_to(attn_mask, layer.shape)
        layer = tf.where(attn_mask == 1, layer, fill_val)

        if self.pooling_strategy == 'cls':
            CLS = self.tokenizer.cls_token_id
            ind = tf.experimental.numpy.nonzero(
                input_tokens['input_ids'] == CLS
            )
            output = tf.gather_nd(layer, tf.stack(ind, axis=1))
        elif self.pooling_strategy == 'mean':
            output = tf.reduce_sum(layer, 1) / tf.reduce_sum(
                tf.cast(attn_mask, tf.float32), 1
            )
        elif self.pooling_strategy == 'max':
            output = tf.reduce_max(layer, 1)
        elif self.pooling_strategy == 'min':
            output = tf.reduce_min(layer, 1)

        return output.numpy()


def _cosine(A_norm_ext, B_norm_ext):

    return A_norm_ext.dot(B_norm_ext).clip(min=0) / 2


def _ext_A(A):
    nA, dim = A.shape
    A_ext = _get_ones(nA, dim * 3)
    A_ext[:, dim: 2 * dim] = A
    A_ext[:, 2 * dim:] = A ** 2
    return A_ext


def _ext_B(B):
    nB, dim = B.shape
    B_ext = _get_ones(dim * 3, nB)
    B_ext[:dim] = (B ** 2).T
    B_ext[dim: 2 * dim] = -2.0 * B.T
    del B
    return B_ext


def _euclidean(A_ext, B_ext):
    sqdist = A_ext.dot(B_ext).clip(min=0)
    return np.sqrt(sqdist)


def _norm(A):
    return A / np.linalg.norm(A, ord=2, axis=1, keepdims=True)


def _get_ones(x, y):
    return np.ones((x, y))


def _expand_vector(vec):
    if not isinstance(vec, np.ndarray):
        vec = np.array(vec)
    if len(vec.shape) == 1:
        vec = np.expand_dims(vec, 0)
    return vec



__copyright__ = "Copyright (c) 2021 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

import numpy as np

from jina.executors.decorators import batching, as_ndarray
from jina.executors.encoders.frameworks import BaseTorchEncoder

class SpacyTextEncoder(BaseTorchEncoder):
    """
    :class:`SpacyTextEncoder` encodes ``Document`` content from a `np.ndarray` (of strings) of length `BatchSize` into
    a `np.ndarray` of shape `Batchsize x EmbeddingDimension`.

    :param lang: pre-trained spaCy language pipeline (model name HashEmbedCNN by default for tok2vec), `en_core_web_sm`
        by default. Allows models `en_core_web_md`, `en_core_web_lg`, `en_core_web_trf`. Refer https://spacy.io/models/en.
    :param use_default_encoder: if True will use parser component,
        otherwise tok2vec implementation will be chosen,
        by default False.
    :param args: Additional positional arguments.
    :param kwargs: Additional positional arguments.
    """

    SPACY_COMPONENTS = [
        'tagger',
        'parser',
        'ner',
        'senter',
        'tok2vec',
        'lemmatizer',
        'attribute_ruler',
    ]
    def __init__(self, lang: str = 'en_core_web_sm', use_default_encoder: bool = False, *args, **kwargs):
        """Set constructor."""
        super().__init__(*args, **kwargs)
        self.lang = lang
        self.use_default_encoder = use_default_encoder

    def post_init(self):
        """Load a model from spacy specified in `lang`. """
        import spacy

        try:
            self.spacy_model = spacy.load(self.lang)
            # Disable everything as we only requires certain pipelines to turned on.
            ignored_components = []
            for comp in self.SPACY_COMPONENTS:
                try:
                    self.spacy_model.disable_pipe(comp)
                except Exception:
                    ignored_components.append(comp)
            self.logger.info(f'Ignoring {ignored_components} pipelines as it does not available on the model package.')
        except IOError:
            self.logger.error(
                f'spaCy model for language {self.lang} can not be found. Please install by referring to the '
                'official page https://spacy.io/usage/models.'
            )
            raise

        if self.use_default_encoder:
            try:
                self.spacy_model.enable_pipe('parser')
            except ValueError:
                self.logger.error(
                    f'Parser for language {self.lang} can not be found. The default sentence encoder requires'
                    'DependencyParser to be trained. Please refer to https://spacy.io/api/tok2vec for more clarity.'
                )
                raise
        else:
            try:
                self.spacy_model.enable_pipe('tok2vec')
            except ValueError:
                self.logger.error(
                    f'TokenToVector is not available for language {self.lang}. Please refer to'
                    'https://github.com/explosion/spaCy/issues/6615 for training your own recognizer.'
                )
                raise

    @batching
    @as_ndarray
    def encode(self, content: 'np.ndarray', *args, **kwargs) -> 'np.ndarray':
        """
        Transform a `np.ndarray` of strings of length `BatchSize` into
        a `np.ndarray` of shape `Batchsize x EmbeddingDimension`.
        Calculate word embeddings by applying spacy's tok2vec component
        model in a sentence wise manner.

        :param content: A `np.ndarray` of strings. Each string is a sentence.
        :param args: Additional positional arguments.
        :param kwargs: Additional positional arguments.
        :return: A `BachSize x EmbeddingSize` numpy `ndarray`.
        """
        embedding = []
        for sent_data in content:
            processed_data = self.spacy_model(str(sent_data))
            result = [self.tensor2array(token.tensor) for token in processed_data]
            embedding.append(result)
        return np.vstack(embedding)

    def tensor2array(self, tensor):
        if isinstance(tensor, np.ndarray):
            return tensor
        return tensor.cpu().numpy() if self.on_gpu else tensor.numpy()



from typing import Tuple, Optional, Iterable

from .. import BaseExecutableDriver, FlatRecursiveMixin
from ...types.sets import MatchSet
from ...types.score import NamedScore

if False:
    from ...types.sets import DocumentSet


class BaseRankDriver(FlatRecursiveMixin, BaseExecutableDriver):
    """Drivers inherited from this Driver will bind :meth:`rank` by default """

    def __init__(
        self, executor: Optional[str] = None, method: str = 'score', *args, **kwargs
    ):
        super().__init__(executor, method, *args, **kwargs)

    @property
    def _exec_match_keys(self):
        """Property to provide backward compatibility to executors relying in `required_keys`
        :return: keys for attribute lookup in matches
        """
        return getattr(
            self.exec, 'match_required_keys', getattr(self.exec, 'required_keys', None)
        )

    @property
    def _exec_query_keys(self):
        """Property to provide backward compatibility to executors relying in `required_keys`

        :return: keys for attribute lookup in matches
        """
        return getattr(
            self.exec, 'query_required_keys', getattr(self.exec, 'required_keys', None)
        )


class Matches2DocRankDriver(BaseRankDriver):
    """This driver is intended to only resort the given matches on the 0 level granularity for a document.
    It gets the scores from a Ranking Executor, which does only change the scores of matches.
    Afterwards, the Matches2DocRankDriver resorts all matches for a document.
    Input-Output ::
        Input:
        document: {granularity: 0, adjacency: k}
            |- matches: {granularity: 0, adjacency: k+1}
        Output:
        document: {granularity: 0, adjacency: k}
            |- matches: {granularity: 0, adjacency: k+1} (Sorted according to scores from Ranker Executor)
    """

    def __init__(
        self,
        reverse: bool = True,
        traversal_paths: Tuple[str] = ('r',),
        *args,
        **kwargs,
    ):
        super().__init__(traversal_paths=traversal_paths, *args, **kwargs)
        self.reverse = reverse

    def _apply_all(self, docs: 'DocumentSet', *args, **kwargs) -> None:
        """

        :param docs: the matches of the ``context_doc``, they are at granularity ``k``
        :param args: not used (kept to maintain interface)
        :param kwargs: not used (kept to maintain interface)

        .. note::
            - This driver will change in place the ordering of ``matches`` of the ``context_doc`.
            - Set the ``traversal_paths`` of this driver such that it traverses along the ``matches`` of the ``chunks`` at the level desired.
        """
        old_scores = []
        queries_metas = []
        matches_metas = []
        for doc in docs:
            query_meta = (
                doc.get_attrs(*self._exec_query_keys) if self._exec_query_keys else None
            )

            matches = doc.matches
            old_match_scores = []
            needs_match_meta = self._exec_match_keys is not None
            match_meta = [] if needs_match_meta else None
            for match in matches:
                old_match_scores.append(match.score.value)
                if needs_match_meta:
                    match_meta.append(match.get_attrs(*self._exec_match_keys))

            # if there are no matches, no need to sort them
            old_scores.append(old_match_scores)
            queries_metas.append(query_meta)
            matches_metas.append(match_meta)

        new_scores = self.exec_fn(old_scores, queries_metas, matches_metas)
        if len(new_scores) != len(docs):
            msg = f'The number of scores {len(new_scores)} does not match the number of queries {len(docs)}'
            self.logger.error(msg)
            raise ValueError(msg)

        for doc, scores in zip(docs, new_scores):
            matches = doc.matches
            if len(doc.matches) != len(scores):
                msg = (
                    f'The number of matches to be scored {len(doc.matches)} do not match the number of scores returned '
                    f'by the ranker {self.exec.__name__} for doc: {doc.id} '
                )
                self.logger.error(msg)
                raise ValueError(msg)
            self._sort_matches_in_place(matches, scores)

    def _sort_matches_in_place(
        self, matches: 'MatchSet', match_scores: Iterable[float]
    ) -> None:
        op_name = self.exec.__class__.__name__
        ref_doc_id = matches._ref_doc.id

        for match, score in zip(matches, match_scores):
            match.score = NamedScore(value=score, op_name=op_name, ref_id=ref_doc_id)

        matches.sort(key=lambda x: x.score.value, reverse=self.reverse)



__copyright__ = "Copyright (c) 2021 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

from jina.executors.encoders.numeric import TransformEncoder


class IncrementalPCAEncoder(TransformEncoder):
    """
    Encodes ``Document`` content from an ndarray in size `B x T` into an ndarray in size `B x D`.

    Where `B` is the batch's size and `T` and `D` are the dimensions pre (`T`)
    and after (`D`) the encoding.
    """


from .. import BaseNdArray
from ....proto import jina_pb2

__all__ = ['BaseDenseNdArray']


class BaseDenseNdArray(BaseNdArray):
    """
    The base class for DenseNdArray.

    Do not use this class directly. Subclass should be used.
    """

    def __init__(self, *args, **kwargs):
        """Set constructor method."""
        super().__init__(*args, **kwargs)
        self.is_sparse = False

    def null_proto(self):
        """Get the new protobuf representation."""
        return jina_pb2.DenseNdArrayProto()



from jina.executors.encoders.numeric import TransformEncoder


class UMAPEncoder(TransformEncoder):
    """
    :class:`UMAPEncoder` encodes ``Document`` content using Uniform Manifold Approximation and Projection Embedding.

    Encodes ``Document`` content from an ndarray of size `B x T` into an ndarray of size `B x D`
    Where `B` is the batch's size and `T` and `D` are the dimensions pre (`T`)
    and after (`D`) the encoding.

    Full code and documentation can be found
    `here <https://github.com/lmcinnes/umap>`_.

    :param model_path: Path to pre-trained model
    :param args: Additional positional arguments
    :param kwargs: Additional keyword arguments
    """
    def __init__(self,
                 model_path: str,
                 *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.model_path = model_path



import base64
import io
import json
import mimetypes
import os
import urllib.parse
import urllib.request
import warnings
from hashlib import blake2b
from typing import Union, Dict, Optional, TypeVar, Any, Tuple, List, Type

import numpy as np
from google.protobuf import json_format
from google.protobuf.field_mask_pb2 import FieldMask

from .converters import png_to_buffer, to_datauri, guess_mime, to_image_blob
from .traversable import Traversable
from ..mixin import ProtoTypeMixin
from ..ndarray.generic import NdArray, BaseSparseNdArray
from ..querylang.queryset.dunderkey import dunder_get
from ..score import NamedScore
from ..sets.chunk import ChunkSet
from ..sets.match import MatchSet
from ...excepts import BadDocType
from ...helper import (
    is_url,
    typename,
    random_identity,
    download_mermaid_url,
)
from ...importer import ImportExtensions
from ...logging import default_logger
from ...proto import jina_pb2

if False:
    from scipy.sparse import coo_matrix

    # fix type-hint complain for sphinx and flake
    from typing import TypeVar
    import numpy as np
    import scipy
    import tensorflow as tf
    import torch

    EmbeddingType = TypeVar(
        'EncodingType',
        np.ndarray,
        scipy.sparse.csr_matrix,
        scipy.sparse.coo_matrix,
        scipy.sparse.bsr_matrix,
        scipy.sparse.csc_matrix,
        torch.sparse_coo_tensor,
        tf.SparseTensor,
    )

    SparseEmbeddingType = TypeVar(
        'SparseEmbeddingType',
        np.ndarray,
        scipy.sparse.csr_matrix,
        scipy.sparse.coo_matrix,
        scipy.sparse.bsr_matrix,
        scipy.sparse.csc_matrix,
        torch.sparse_coo_tensor,
        tf.SparseTensor,
    )

__all__ = ['Document', 'DocumentContentType', 'DocumentSourceType']
DIGEST_SIZE = 8

DocumentContentType = TypeVar('DocumentContentType', bytes, str, np.ndarray)
DocumentSourceType = TypeVar(
    'DocumentSourceType', jina_pb2.DocumentProto, bytes, str, Dict
)

_document_fields = set(
    list(jina_pb2.DocumentProto().DESCRIPTOR.fields_by_camelcase_name)
    + list(jina_pb2.DocumentProto().DESCRIPTOR.fields_by_name)
)

_all_mime_types = set(mimetypes.types_map.values())


class Document(ProtoTypeMixin, Traversable):
    """
    :class:`Document` is one of the **primitive data type** in Jina.

    It offers a Pythonic interface to allow users access and manipulate
    :class:`jina.jina_pb2.DocumentProto` object without working with Protobuf itself.

    To create a :class:`Document` object, simply:

        .. highlight:: python
        .. code-block:: python

            from jina import Document
            d = Document()
            d.text = 'abc'

    Jina requires each Document to have a string id. You can set a custom one,
    or if non has been set a random one will be assigned.

    Or you can use :class:`Document` as a context manager:

        .. highlight:: python
        .. code-block:: python

            with Document() as d:
                d.text = 'hello'

            assert d.id  # now `id` has value

    To access and modify the content of the document, you can use :attr:`text`, :attr:`blob`, and :attr:`buffer`.
    Each property is implemented with proper setter, to improve the integrity and user experience. For example,
    assigning ``doc.blob`` or ``doc.embedding`` can be simply done via:

        .. highlight:: python
        .. code-block:: python

            import numpy as np

            # to set as content
            d.content = np.random.random([10, 5])

            # to set as embedding
            d.embedding = np.random.random([10, 5])

    MIME type is auto set/guessed when setting :attr:`content` and :attr:`uri`

    :class:`Document` also provides multiple way to build from existing Document. You can build :class:`Document`
    from ``jina_pb2.DocumentProto``, ``bytes``, ``str``, and ``Dict``. You can also use it as view (i.e.
    weak reference when building from an existing ``jina_pb2.DocumentProto``). For example,

        .. highlight:: python
        .. code-block:: python

            a = DocumentProto()
            b = Document(a, copy=False)
            a.text = 'hello'
            assert b.text == 'hello'

    You can leverage the :meth:`convert_a_to_b` interface to convert between content forms.

    """

    def __init__(
        self,
        document: Optional[DocumentSourceType] = None,
        field_resolver: Dict[str, str] = None,
        copy: bool = False,
        **kwargs,
    ):
        """
        :param document: the document to construct from. If ``bytes`` is given
                then deserialize a :class:`DocumentProto`; ``dict`` is given then
                parse a :class:`DocumentProto` from it; ``str`` is given, then consider
                it as a JSON string and parse a :class:`DocumentProto` from it; finally,
                one can also give `DocumentProto` directly, then depending on the ``copy``,
                it builds a view or a copy from it.
        :param copy: when ``document`` is given as a :class:`DocumentProto` object, build a
                view (i.e. weak reference) from it or a deep copy from it.
        :param field_resolver: a map from field names defined in ``document`` (JSON, dict) to the field
                names defined in Protobuf. This is only used when the given ``document`` is
                a JSON string or a Python dict.
        :param kwargs: other parameters to be set _after_ the document is constructed

        .. note::

            When ``document`` is a JSON string or Python dictionary object, the constructor will only map the values
            from known fields defined in Protobuf, all unknown fields are mapped to ``document.tags``. For example,

            .. highlight:: python
            .. code-block:: python

                d = Document({'id': '123', 'hello': 'world', 'tags': {'good': 'bye'}})

                assert d.id == '123'  # true
                assert d.tags['hello'] == 'world'  # true
                assert d.tags['good'] == 'bye'  # true
        """
        self._pb_body = jina_pb2.DocumentProto()
        try:
            if isinstance(document, jina_pb2.DocumentProto):
                if copy:
                    self._pb_body.CopyFrom(document)
                else:
                    self._pb_body = document
            elif isinstance(document, (dict, str)):
                if isinstance(document, str):
                    document = json.loads(document)

                if field_resolver:
                    document = {
                        field_resolver.get(k, k): v for k, v in document.items()
                    }

                user_fields = set(document.keys())
                if _document_fields.issuperset(user_fields):
                    json_format.ParseDict(document, self._pb_body)
                else:
                    _intersect = _document_fields.intersection(user_fields)
                    _remainder = user_fields.difference(_intersect)
                    if _intersect:
                        json_format.ParseDict(
                            {k: document[k] for k in _intersect}, self._pb_body
                        )
                    if _remainder:
                        self._pb_body.tags.update({k: document[k] for k in _remainder})
            elif isinstance(document, bytes):
                # directly parsing from binary string gives large false-positive
                # fortunately protobuf throws a warning when the parsing seems go wrong
                # the context manager below converts this warning into exception and throw it
                # properly
                with warnings.catch_warnings():
                    warnings.filterwarnings(
                        'error', 'Unexpected end-group tag', category=RuntimeWarning
                    )
                    try:
                        self._pb_body.ParseFromString(document)
                    except RuntimeWarning as ex:
                        raise BadDocType(
                            f'fail to construct a document from {document}'
                        ) from ex
            elif isinstance(document, Document):
                if copy:
                    self._pb_body.CopyFrom(document.proto)
                else:
                    self._pb_body = document.proto
            elif document is not None:
                # note ``None`` is not considered as a bad type
                raise ValueError(f'{typename(document)} is not recognizable')
        except Exception as ex:
            raise BadDocType(
                f'fail to construct a document from {document}, '
                f'if you are trying to set the content '
                f'you may use "Document(content=your_content)"'
            ) from ex

        if self._pb_body.id is None or not self._pb_body.id:
            self.id = random_identity(use_uuid1=True)

        self.set_attrs(**kwargs)
        self._mermaid_id = random_identity()  #: for mermaid visualize id

    @property
    def siblings(self) -> int:
        """
        The number of siblings of the :class:``Document``

        .. # noqa: DAR201
        :getter: number of siblings
        :setter: number of siblings
        :type: int
        """
        return self._pb_body.siblings

    @siblings.setter
    def siblings(self, value: int):
        self._pb_body.siblings = value

    @property
    def weight(self) -> float:
        """
        :return: the weight of the document
        """
        return self._pb_body.weight

    @weight.setter
    def weight(self, value: float):
        """
        Set the weight of the document.

        :param value: the float weight of the document.
        """
        self._pb_body.weight = value

    @property
    def modality(self) -> str:
        """
        :return: the modality of the document."""
        return self._pb_body.modality

    @modality.setter
    def modality(self, value: str):
        """Set the modality of the document.

        :param value: The modality of the document
        """
        self._pb_body.modality = value

    @property
    def content_hash(self):
        """Get the content hash of the document.

        :return: the content_hash from the proto
        """
        return self._pb_body.content_hash

    @staticmethod
    def _update(
        source: 'Document',
        destination: 'Document',
        exclude_fields: Optional[Tuple[str]] = None,
        include_fields: Optional[Tuple[str]] = None,
        replace_message_field: bool = True,
        replace_repeated_field: bool = True,
    ) -> None:
        """Merge fields specified in ``include_fields`` or ``exclude_fields`` from source to destination.

        :param source: source :class:`Document` object.
        :param destination: the destination :class:`Document` object to be merged into.
        :param exclude_fields: a tuple of field names that excluded from destination document
        :param include_fields: a tuple of field names that included from source document
        :param replace_message_field: Replace message field if True. Merge message
                  field if False.
        :param replace_repeated_field: Replace repeated field if True. Append
                  elements of repeated field if False.

        .. note::
            *. if neither ``exclude_fields`` nor ``include_fields`` is given,
                then destination is overridden by the source completely.
            *. ``destination`` will be modified in place, ``source`` will be unchanged
        """

        if not include_fields and not exclude_fields:
            # same behavior as copy
            destination.CopyFrom(source)
        elif include_fields is not None and exclude_fields is None:
            FieldMask(paths=include_fields).MergeMessage(
                source.proto,
                destination.proto,
                replace_message_field=replace_message_field,
                replace_repeated_field=replace_repeated_field,
            )
        elif exclude_fields is not None:
            empty_doc = jina_pb2.DocumentProto()

            _dest = jina_pb2.DocumentProto()
            # backup exclude fields in destination
            FieldMask(paths=exclude_fields).MergeMessage(
                destination.proto,
                _dest,
                replace_repeated_field=True,
                replace_message_field=True,
            )

            if include_fields is None:
                # override dest with src
                destination.CopyFrom(source)
            else:
                # only update include fields
                FieldMask(paths=include_fields).MergeMessage(
                    source.proto,
                    destination.proto,
                    replace_message_field=replace_message_field,
                    replace_repeated_field=replace_repeated_field,
                )

            # clear the exclude fields
            FieldMask(paths=exclude_fields).MergeMessage(
                empty_doc,
                destination.proto,
                replace_repeated_field=True,
                replace_message_field=True,
            )

            # recover exclude fields
            destination.proto.MergeFrom(_dest)

    def update(
        self,
        source: 'Document',
        exclude_fields: Optional[Tuple[str, ...]] = None,
        include_fields: Optional[Tuple[str, ...]] = None,
    ) -> None:
        """Updates fields specified in ``include_fields`` from the source to current Document.

        :param source: source :class:`Document` object.
        :param exclude_fields: a tuple of field names that excluded from the current document,
                when not given the non-empty fields of the current document is considered as ``exclude_fields``
        :param include_fields: a tuple of field names that included from the source document

        .. note::
            *. ``destination`` will be modified in place, ``source`` will be unchanged
        """
        if (include_fields and not isinstance(include_fields, tuple)) or (
            exclude_fields and not isinstance(exclude_fields, tuple)
        ):
            raise TypeError('include_fields and exclude_fields must be tuple of str')

        if exclude_fields is None:
            if include_fields:
                exclude_fields = tuple(
                    f for f in self.non_empty_fields if f not in include_fields
                )
            else:
                exclude_fields = self.non_empty_fields

        if include_fields and exclude_fields:
            _intersect = set(include_fields).intersection(exclude_fields)
            if _intersect:
                raise ValueError(
                    f'{_intersect} is in both `include_fields` and `exclude_fields`'
                )

        self._update(
            source,
            self,
            exclude_fields=exclude_fields,
            include_fields=include_fields,
            replace_message_field=True,
            replace_repeated_field=True,
        )

    def update_content_hash(
        self,
        exclude_fields: Optional[Tuple[str]] = (
            'id',
            'chunks',
            'matches',
            'content_hash',
            'parent_id',
        ),
        include_fields: Optional[Tuple[str]] = None,
    ) -> None:
        """Update the document hash according to its content.

        :param exclude_fields: a tuple of field names that excluded when computing content hash
        :param include_fields: a tuple of field names that included when computing content hash

        .. note::
            "exclude_fields" and "include_fields" are mutually exclusive, use one only
        """
        masked_d = jina_pb2.DocumentProto()
        masked_d.CopyFrom(self._pb_body)
        empty_doc = jina_pb2.DocumentProto()
        if include_fields and exclude_fields:
            raise ValueError(
                '"exclude_fields" and "exclude_fields" are mutually exclusive, use one only'
            )

        if include_fields is not None:
            FieldMask(paths=include_fields).MergeMessage(masked_d, empty_doc)
            masked_d = empty_doc
        elif exclude_fields is not None:
            FieldMask(paths=exclude_fields).MergeMessage(
                empty_doc, masked_d, replace_repeated_field=True
            )

        self._pb_body.content_hash = blake2b(
            masked_d.SerializeToString(), digest_size=DIGEST_SIZE
        ).hexdigest()

    @property
    def id(self) -> str:
        """The document id in hex string, for non-binary environment such as HTTP, CLI, HTML and also human-readable.
        it will be used as the major view.

        :return: the id from the proto
        """
        return self._pb_body.id

    @property
    def parent_id(self) -> str:
        """The document's parent id in hex string, for non-binary environment such as HTTP, CLI, HTML and also human-readable.
        it will be used as the major view.

        :return: the parent id from the proto
        """
        return self._pb_body.parent_id

    @id.setter
    def id(self, value: Union[bytes, str, int]):
        """Set document id to a string value.

        :param value: id as bytes, int or str
        """
        self._pb_body.id = str(value)

    @parent_id.setter
    def parent_id(self, value: Union[bytes, str, int]):
        """Set document's parent id to a string value.

        :param value: id as bytes, int or str
        """
        self._pb_body.parent_id = str(value)

    @property
    def blob(self) -> 'np.ndarray':
        """Return ``blob``, one of the content form of a Document.

        .. note::
            Use :attr:`content` to return the content of a Document

        :return: the blob content from the proto
        """
        return NdArray(self._pb_body.blob).value

    @blob.setter
    def blob(self, value: Union['np.ndarray', 'jina_pb2.NdArrayProto', 'NdArray']):
        """Set the `blob` to :param:`value`.

        :param value: the array value to set the blob
        """
        self._update_ndarray('blob', value)

    @property
    def embedding(self) -> 'EmbeddingType':
        """Return ``embedding`` of the content of a Document.

        :return: the embedding from the proto
        """
        return NdArray(self._pb_body.embedding).value

    def get_sparse_embedding(
        self, sparse_ndarray_cls_type: Type[BaseSparseNdArray], **kwargs
    ) -> 'SparseEmbeddingType':
        """Return ``embedding`` of the content of a Document as an sparse array.

        :param sparse_ndarray_cls_type: Sparse class type, such as `SparseNdArray`.
        :param kwargs: Additional key value argument, for `scipy` backend, we need to set
            the keyword `sp_format` as one of the scipy supported sparse format, such as `coo`
            or `csr`.
        :return: the embedding from the proto as an sparse array
        """
        return NdArray(
            self._pb_body.embedding,
            sparse_cls=sparse_ndarray_cls_type,
            is_sparse=True,
            **kwargs,
        ).value

    @embedding.setter
    def embedding(self, value: Union['np.ndarray', 'jina_pb2.NdArrayProto', 'NdArray']):
        """Set the ``embedding`` of the content of a Document.

        :param value: the array value to set the embedding
        """
        self._update_ndarray('embedding', value)

    def _update_sparse_ndarray(self, k, v, sparse_cls):
        NdArray(
            is_sparse=True,
            sparse_cls=sparse_cls,
            proto=getattr(self._pb_body, k),
        ).value = v

    def _check_installed_array_packages(self):
        from ... import JINA_GLOBAL

        if JINA_GLOBAL.scipy_installed is None:
            JINA_GLOBAL.scipy_installed = False
            with ImportExtensions(required=False, pkg_name='scipy'):
                import scipy

                JINA_GLOBAL.scipy_installed = True

        if JINA_GLOBAL.tensorflow_installed is None:
            JINA_GLOBAL.tensorflow_installed = False
            with ImportExtensions(required=False, pkg_name='tensorflow'):
                import tensorflow

                JINA_GLOBAL.tensorflow_installed = True

        if JINA_GLOBAL.torch_installed is None:
            JINA_GLOBAL.torch_installed = False
            with ImportExtensions(required=False, pkg_name='torch'):
                import torch

                JINA_GLOBAL.torch_installed = True

    def _update_if_sparse(self, k, v):

        from ... import JINA_GLOBAL

        v_valid_sparse_type = False
        self._check_installed_array_packages()

        if JINA_GLOBAL.scipy_installed:
            import scipy

            if scipy.sparse.issparse(v):
                from ..ndarray.sparse.scipy import SparseNdArray

                self._update_sparse_ndarray(k=k, v=v, sparse_cls=SparseNdArray)
                v_valid_sparse_type = True

        if JINA_GLOBAL.tensorflow_installed:
            import tensorflow

            if isinstance(v, tensorflow.SparseTensor):
                from ..ndarray.sparse.tensorflow import SparseNdArray

                self._update_sparse_ndarray(k=k, v=v, sparse_cls=SparseNdArray)
                v_valid_sparse_type = True

        if JINA_GLOBAL.torch_installed:
            import torch

            if isinstance(v, torch.Tensor) and v.is_sparse:
                from ..ndarray.sparse.pytorch import SparseNdArray

                self._update_sparse_ndarray(k=k, v=v, sparse_cls=SparseNdArray)
                v_valid_sparse_type = True

        return v_valid_sparse_type

    def _update_ndarray(self, k, v):
        if isinstance(v, jina_pb2.NdArrayProto):
            getattr(self._pb_body, k).CopyFrom(v)
        elif isinstance(v, np.ndarray):
            NdArray(getattr(self._pb_body, k)).value = v
        elif isinstance(v, NdArray):
            NdArray(getattr(self._pb_body, k)).is_sparse = v.is_sparse
            NdArray(getattr(self._pb_body, k)).value = v.value

        else:
            v_valid_sparse_type = self._update_if_sparse(k, v)

            if not v_valid_sparse_type:
                raise TypeError(f'{k} is in unsupported type {typename(v)}')

    @property
    def matches(self) -> 'MatchSet':
        """Get all matches of the current document.

        :return: the set of matches attached to this document
        """
        return MatchSet(self._pb_body.matches, reference_doc=self)

    @property
    def chunks(self) -> 'ChunkSet':
        """Get all chunks of the current document.

        :return: the set of chunks of this document
        """
        return ChunkSet(self._pb_body.chunks, reference_doc=self)

    def __getattr__(self, item):
        if hasattr(self._pb_body, item):
            value = getattr(self._pb_body, item)
        else:
            value = dunder_get(self._pb_body, item)
        return value

    def set_attrs(self, **kwargs):
        """Bulk update Document fields with key-value specified in kwargs

        .. seealso::
            :meth:`get_attrs` for bulk get attributes

        :param kwargs: the keyword arguments to set the values, where the keys are the fields to set
        """
        for k, v in kwargs.items():
            if isinstance(v, (list, tuple)):
                if k == 'chunks':
                    self.chunks.extend(v)
                elif k == 'matches':
                    self.matches.extend(v)
                else:
                    self._pb_body.ClearField(k)
                    getattr(self._pb_body, k).extend(v)
            elif isinstance(v, dict):
                self._pb_body.ClearField(k)
                getattr(self._pb_body, k).update(v)
            else:
                if (
                    hasattr(Document, k)
                    and isinstance(getattr(Document, k), property)
                    and getattr(Document, k).fset
                ):
                    # if class property has a setter
                    setattr(self, k, v)
                elif hasattr(self._pb_body, k):
                    # no property setter, but proto has this attribute so fallback to proto
                    setattr(self._pb_body, k, v)
                else:
                    raise AttributeError(f'{k} is not recognized')

    def get_attrs(self, *args) -> Dict[str, Any]:
        """Bulk fetch Document fields and return a dict of the key-value pairs

        .. seealso::
            :meth:`update` for bulk set/update attributes

        .. note::
            Arguments will be extracted using `dunder_get`
            .. highlight:: python
            .. code-block:: python

                d = Document({'id': '123', 'hello': 'world', 'tags': {'id': 'external_id', 'good': 'bye'}})

                assert d.id == '123'  # true
                assert d.tags['hello'] == 'world' # true
                assert d.tags['good'] == 'bye' # true
                assert d.tags['id'] == 'external_id' # true

                res = d.get_attrs(*['id', 'tags__hello', 'tags__good', 'tags__id'])

                assert res['id'] == '123' # true
                assert res['tags__hello'] == 'world' # true
                assert res['tags__good'] == 'bye' # true
                assert res['tags__id'] == 'external_id' # true

        :param args: the variable length values to extract from the document
        :return: a dictionary mapping the fields in `:param:args` to the actual attributes of this document
        """

        ret = {}
        for k in args:
            try:
                value = getattr(self, k)

                if value is None:
                    raise ValueError

                ret[k] = value
            except (AttributeError, ValueError):
                default_logger.warning(
                    f'Could not get attribute `{typename(self)}.{k}`, returning `None`'
                )
                ret[k] = None
        return ret

    def get_attrs_values(self, *args) -> List[Any]:
        """Bulk fetch Document fields and return a list of the values of these fields

        .. note::
            Arguments will be extracted using `dunder_get`
            .. highlight:: python
            .. code-block:: python

                d = Document({'id': '123', 'hello': 'world', 'tags': {'id': 'external_id', 'good': 'bye'}})

                assert d.id == '123'  # true
                assert d.tags['hello'] == 'world' # true
                assert d.tags['good'] == 'bye' # true
                assert d.tags['id'] == 'external_id' # true

                res = d.get_attrs_values(*['id', 'tags__hello', 'tags__good', 'tags__id'])

                assert res == ['123', 'world', 'bye', 'external_id']

        :param args: the variable length values to extract from the document
        :return: a list with the attributes of this document ordered as the args
        """

        ret = []
        for k in args:
            try:
                value = getattr(self, k)

                if value is None:
                    raise ValueError

                ret.append(value)
            except (AttributeError, ValueError):
                default_logger.warning(
                    f'Could not get attribute `{typename(self)}.{k}`, returning `None`'
                )
                ret.append(None)

        return ret

    @property
    def buffer(self) -> bytes:
        """Return ``buffer``, one of the content form of a Document.

        .. note::
            Use :attr:`content` to return the content of a Document

        :return: the buffer bytes from this document
        """
        return self._pb_body.buffer

    @buffer.setter
    def buffer(self, value: bytes):
        """Set the ``buffer`` to :param:`value`.

        :param value: the bytes value to set the buffer
        """
        self._pb_body.buffer = value
        if value and not self._pb_body.mime_type:
            with ImportExtensions(
                required=False,
                pkg_name='python-magic',
                help_text=f'can not sniff the MIME type '
                f'MIME sniffing requires brew install '
                f'libmagic (Mac)/ apt-get install libmagic1 (Linux)',
            ):
                import magic

                self._pb_body.mime_type = magic.from_buffer(value, mime=True)

    @property
    def text(self):
        """Return ``text``, one of the content form of a Document.

        .. note::
            Use :attr:`content` to return the content of a Document

        :return: the text from this document content
        """
        return self._pb_body.text

    @text.setter
    def text(self, value: str):
        """Set the `text` to :param:`value`

        :param value: the text value to set as content
        """
        self._pb_body.text = value
        self.mime_type = 'text/plain'

    @property
    def uri(self) -> str:
        """Return the URI of the document.

        :return: the uri from this document proto
        """
        return self._pb_body.uri

    @uri.setter
    def uri(self, value: str):
        """Set the URI of the document.

        .. note::
            :attr:`mime_type` will be updated accordingly

        :param value: acceptable URI/URL, raise ``ValueError`` when it is not a valid URI
        """
        scheme = urllib.parse.urlparse(value).scheme
        if (
            (scheme in {'http', 'https'} and is_url(value))
            or (scheme in {'data'})
            or os.path.exists(value)
            or os.access(os.path.dirname(value), os.W_OK)
        ):
            self._pb_body.uri = value
            self.mime_type = guess_mime(value)
        else:
            raise ValueError(f'{value} is not a valid URI')

    @property
    def mime_type(self) -> str:
        """Get MIME type of the document

        :return: the mime_type from this document proto
        """
        return self._pb_body.mime_type

    @mime_type.setter
    def mime_type(self, value: str):
        """Set MIME type of the document

        :param value: the acceptable MIME type, raise ``ValueError`` when MIME type is not
                recognizable.
        """
        if value in _all_mime_types:
            self._pb_body.mime_type = value
        elif value:
            # given but not recognizable, do best guess
            r = mimetypes.guess_type(f'*.{value}')[0]
            if r:
                self._pb_body.mime_type = r
            else:
                self._pb_body.mime_type = value

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.update_content_hash()

    @property
    def content_type(self) -> str:
        """Return the content type of the document, possible values: text, blob, buffer

        :return: the type of content present in this document proto
        """
        return self._pb_body.WhichOneof('content')

    @property
    def content(self) -> DocumentContentType:
        """Return the content of the document. It checks whichever field among :attr:`blob`, :attr:`text`,
        :attr:`buffer` has value and return it.

        .. seealso::
            :attr:`blob`, :attr:`buffer`, :attr:`text`

        :return: the value of the content depending on `:meth:`content_type`
        """
        attr = self.content_type
        if attr:
            return getattr(self, attr)

    @content.setter
    def content(self, value: DocumentContentType):
        """Set the content of the document. It assigns the value to field with the right type.

        .. seealso::
            :attr:`blob`, :attr:`buffer`, :attr:`text`

        :param value: the value from which to set the content of the Document
        """
        if isinstance(value, bytes):
            self.buffer = value
        elif isinstance(value, str):
            # TODO(Han): this implicit fallback is too much but that's
            #  how the original _generate function implement. And a lot of
            #  tests depend on this logic. Stay in this
            #  way to keep all tests passing until I got time to refactor this part
            try:
                self.uri = value
            except ValueError:
                self.text = value
        elif isinstance(value, np.ndarray):
            self.blob = value
        else:
            # ``None`` is also considered as bad type
            raise TypeError(f'{typename(value)} is not recognizable')

    @property
    def granularity(self):
        """Return the granularity of the document.

        :return: the granularity from this document proto
        """
        return self._pb_body.granularity

    @granularity.setter
    def granularity(self, value: int):
        """Set the granularity of the document.

        :param value: the value of the granularity to be set
        """
        self._pb_body.granularity = value

    @property
    def adjacency(self):
        """Return the adjacency of the document.

        :return: the adjacency from this document proto
        """
        return self._pb_body.adjacency

    @adjacency.setter
    def adjacency(self, value: int):
        """Set the adjacency of the document.

        :param value: the value of the adjacency to be set
        """
        self._pb_body.adjacency = value

    @property
    def score(self):
        """Return the score of the document.

        :return: the score attached to this document as `:class:NamedScore`
        """
        return NamedScore(self._pb_body.score)

    @score.setter
    def score(self, value: Union[jina_pb2.NamedScoreProto, NamedScore]):
        """Set the score of the document.

        :param value: the value to set the score of the Document from
        """
        if isinstance(value, jina_pb2.NamedScoreProto):
            self._pb_body.score.CopyFrom(value)
        elif isinstance(value, NamedScore):
            self._pb_body.score.CopyFrom(value._pb_body)
        else:
            raise TypeError(f'score is in unsupported type {typename(value)}')

    def convert_buffer_to_blob(self, **kwargs):
        """Assuming the :attr:`buffer` is a _valid_ buffer of Numpy ndarray,
        set :attr:`blob` accordingly.

        :param kwargs: reserved for maximum compatibility when using with ConvertDriver

        .. note::
            One can only recover values not shape information from pure buffer.
        """
        self.blob = np.frombuffer(self.buffer)

    def convert_buffer_image_to_blob(self, color_axis: int = -1, **kwargs):
        """Convert an image buffer to blob

        :param color_axis: the axis id of the color channel, ``-1`` indicates the color channel info at the last axis
        :param kwargs: reserved for maximum compatibility when using with ConvertDriver
        """
        self.blob = to_image_blob(io.BytesIO(self.buffer), color_axis)

    def convert_blob_to_uri(
        self, width: int, height: int, resize_method: str = 'BILINEAR', **kwargs
    ):
        """Assuming :attr:`blob` is a _valid_ image, set :attr:`uri` accordingly
        :param width: the width of the blob
        :param height: the height of the blob
        :param resize_method: the resize method name
        :param kwargs: reserved for maximum compatibility when using with ConvertDriver
        """
        png_bytes = png_to_buffer(self.blob, width, height, resize_method)
        self.uri = 'data:image/png;base64,' + base64.b64encode(png_bytes).decode()

    def convert_uri_to_blob(
        self, color_axis: int = -1, uri_prefix: Optional[str] = None, **kwargs
    ):
        """Convert uri to blob

        :param color_axis: the axis id of the color channel, ``-1`` indicates the color channel info at the last axis
        :param uri_prefix: the prefix of the uri
        :param kwargs: reserved for maximum compatibility when using with ConvertDriver
        """
        self.blob = to_image_blob(
            (uri_prefix + self.uri) if uri_prefix else self.uri, color_axis
        )

    def convert_data_uri_to_blob(self, color_axis: int = -1, **kwargs):
        """Convert data URI to image blob

        :param color_axis: the axis id of the color channel, ``-1`` indicates the color channel info at the last axis
        :param kwargs: reserved for maximum compatibility when using with ConvertDriver
        """
        req = urllib.request.Request(self.uri, headers={'User-Agent': 'Mozilla/5.0'})
        with urllib.request.urlopen(req) as fp:
            buffer = fp.read()
        self.blob = to_image_blob(io.BytesIO(buffer), color_axis)

    def convert_uri_to_buffer(self, **kwargs):
        """Convert uri to buffer
        Internally it downloads from the URI and set :attr:`buffer`.

        :param kwargs: reserved for maximum compatibility when using with ConvertDriver

        """
        if urllib.parse.urlparse(self.uri).scheme in {'http', 'https', 'data'}:
            req = urllib.request.Request(
                self.uri, headers={'User-Agent': 'Mozilla/5.0'}
            )
            with urllib.request.urlopen(req) as fp:
                self.buffer = fp.read()
        elif os.path.exists(self.uri):
            with open(self.uri, 'rb') as fp:
                self.buffer = fp.read()
        else:
            raise FileNotFoundError(f'{self.uri} is not a URL or a valid local path')

    def convert_uri_to_data_uri(
        self, charset: str = 'utf-8', base64: bool = False, **kwargs
    ):
        """Convert uri to data uri.
        Internally it reads uri into buffer and convert it to data uri

        :param charset: charset may be any character set registered with IANA
        :param base64: used to encode arbitrary octet sequences into a form that satisfies the rules of 7bit. Designed to be efficient for non-text 8 bit and binary data. Sometimes used for text data that frequently uses non-US-ASCII characters.
        :param kwargs: reserved for maximum compatibility when using with ConvertDriver
        """
        self.convert_uri_to_buffer()
        self.uri = to_datauri(self.mime_type, self.buffer, charset, base64, binary=True)

    def convert_buffer_to_uri(
        self, charset: str = 'utf-8', base64: bool = False, **kwargs
    ):
        """Convert buffer to data uri.
        Internally it first reads into buffer and then converts it to data URI.

        :param charset: charset may be any character set registered with IANA
        :param base64: used to encode arbitrary octet sequences into a form that satisfies the rules of 7bit.
            Designed to be efficient for non-text 8 bit and binary data. Sometimes used for text data that
            frequently uses non-US-ASCII characters.
        :param kwargs: reserved for maximum compatibility when using with ConvertDriver
        """

        if not self.mime_type:
            raise ValueError(
                f'{self.mime_type} is unset, can not convert it to data uri'
            )

        self.uri = to_datauri(self.mime_type, self.buffer, charset, base64, binary=True)

    def convert_text_to_uri(
        self, charset: str = 'utf-8', base64: bool = False, **kwargs
    ):
        """Convert text to data uri.

        :param charset: charset may be any character set registered with IANA
        :param base64: used to encode arbitrary octet sequences into a form that satisfies the rules of 7bit.
            Designed to be efficient for non-text 8 bit and binary data.
            Sometimes used for text data that frequently uses non-US-ASCII characters.
        :param kwargs: reserved for maximum compatibility when using with ConvertDriver
        """

        self.uri = to_datauri(self.mime_type, self.text, charset, base64, binary=False)

    def convert_uri_to_text(self, **kwargs):
        """Assuming URI is text, convert it to text

        :param kwargs: reserved for maximum compatibility when using with ConvertDriver
        """
        self.convert_uri_to_buffer()
        self.text = self.buffer.decode()

    def convert_content_to_uri(self, **kwargs):
        """Convert content in URI with best effort

        :param kwargs: reserved for maximum compatibility when using with ConvertDriver
        """
        if self.text:
            self.convert_text_to_uri()
        elif self.buffer:
            self.convert_buffer_to_uri()
        elif self.content_type:
            raise NotImplementedError

    def MergeFrom(self, doc: 'Document'):
        """Merge the content of target :param:doc into current document.

        :param doc: the document to merge from
        """
        self._pb_body.MergeFrom(doc.proto)

    def CopyFrom(self, doc: 'Document'):
        """Copy the content of target :param:doc into current document.

        :param doc: the document to copy from
        """
        self._pb_body.CopyFrom(doc.proto)

    def __mermaid_str__(self):
        results = []
        from google.protobuf.json_format import MessageToDict

        content = MessageToDict(self._pb_body, preserving_proto_field_name=True)

        _id = f'{self._mermaid_id[:3]}~Document~'

        for idx, c in enumerate(self.chunks):
            results.append(
                f'{_id} --> "{idx + 1}/{len(self.chunks)}" {c._mermaid_id[:3]}~Document~: chunks'
            )
            results.append(c.__mermaid_str__())

        for idx, c in enumerate(self.matches):
            results.append(
                f'{_id} ..> "{idx + 1}/{len(self.matches)}" {c._mermaid_id[:3]}~Document~: matches'
            )
            results.append(c.__mermaid_str__())
        if 'chunks' in content:
            content.pop('chunks')
        if 'matches' in content:
            content.pop('matches')
        if content:
            results.append(f'class {_id}{{')
            for k, v in content.items():
                if isinstance(v, (str, int, float, bytes)):
                    results.append(f'+{k} {str(v)[:10]}')
                else:
                    results.append(f'+{k}({type(getattr(self, k, v))})')
            results.append('}')

        return '\n'.join(results)

    def _mermaid_to_url(self, img_type: str) -> str:
        """
        Rendering the current flow as a url points to a SVG, it needs internet connection

        :param img_type: the type of image to be generated
        :return: the url pointing to a SVG
        """
        if img_type == 'jpg':
            img_type = 'img'

        mermaid_str = (
            """
    %%{init: {'theme': 'base', 'themeVariables': { 'primaryColor': '#FFC666'}}}%%
    classDiagram
    
            """
            + self.__mermaid_str__()
        )

        encoded_str = base64.b64encode(bytes(mermaid_str.strip(), 'utf-8')).decode(
            'utf-8'
        )

        return f'https://mermaid.ink/{img_type}/{encoded_str}'

    def _ipython_display_(self):
        """Displays the object in IPython as a side effect"""
        self.plot(inline_display=True)

    def plot(self, output: Optional[str] = None, inline_display: bool = False) -> None:
        """
        Visualize the Document recursively.

        :param output: a filename specifying the name of the image to be created,
                    the suffix svg/jpg determines the file type of the output image
        :param inline_display: show image directly inside the Jupyter Notebook
        """
        image_type = 'svg'
        if output and output.endswith('jpg'):
            image_type = 'jpg'

        url = self._mermaid_to_url(image_type)
        showed = False
        if inline_display:
            try:
                from IPython.display import display, Image

                display(Image(url=url))
                showed = True
            except:
                # no need to panic users
                pass

        if output:
            download_mermaid_url(url, output)
        elif not showed:
            from jina.logging import default_logger

            default_logger.info(f'Document visualization: {url}')

    @property
    def non_empty_fields(self) -> Tuple[str]:
        """Return the set fields of the current document that are not empty

        :return: the tuple of non-empty fields
        """
        return tuple(field[0].name for field in self.ListFields())

    @property
    def raw(self) -> 'Document':
        """Return self as a document object.

        :return: this Document
        """
        return self

    @staticmethod
    def get_all_attributes() -> List[str]:
        """Return all attributes supported by the Document, which can be accessed by ``doc.attribute``

        :return: a list of attributes in string.
        """
        import inspect

        support_keys = list(jina_pb2.DocumentProto().DESCRIPTOR.fields_by_name)

        support_keys += [
            name
            for (name, value) in inspect.getmembers(
                Document, lambda x: isinstance(x, property)
            )
        ]
        return list(set(support_keys))



from . import Message
from ..request import Request
from ...proto import jina_pb2

_available_commands = dict(
    jina_pb2.RequestProto.DumpRequestProto.DESCRIPTOR.enum_values_by_name
)

__all__ = ['DumpMessage']


class DumpMessage(Message):
    """
    Class of the protobuf message.

    :param path: the path to which to dump
    :param shards: the nr of shards to which to dump
    :param pod_name: Name of the current pod.
    :param identity: The identity of the current pod
    :param args: Additional positional arguments which are just used for the parent initialization
    :param kwargs: Additional keyword arguments which are just used for the parent initialization
    """

    def __init__(
        self,
        path: str,
        shards: int,
        pod_name: str = 'ctl',
        identity: str = '',
        *args,
        **kwargs
    ):
        req = Request(jina_pb2.RequestProto())
        req.dump.path = path
        req.dump.shards = shards
        super().__init__(
            None, req, pod_name=pod_name, identity=identity, *args, **kwargs
        )
        req.request_type = 'dump'



__copyright__ = "Copyright (c) 2020 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

import sys
import time
from collections import defaultdict
from functools import wraps
from typing import Optional

from ..importer import ImportExtensions
from ..helper import colored, get_readable_size, get_readable_time

if False:
    # fix type-hint complain for sphinx and flake
    from ..logging import JinaLogger


def used_memory(unit: int = 1024 * 1024 * 1024) -> float:
    """
    Get the memory usage of the current process.

    :param unit: Unit of the memory, default in Gigabytes.
    :return: Memory usage of the current process.
    """
    with ImportExtensions(required=False):
        import resource

        return resource.getrusage(resource.RUSAGE_SELF).ru_maxrss / unit

    from . import default_logger

    default_logger.error(
        'module "resource" can not be found and you are likely running it on Windows, '
        'i will return 0'
    )
    return 0


def used_memory_readable() -> str:
    """
    Get the memory usage of the current process in a human-readable format.

    :return: Memory usage of the current process.
    """
    return get_readable_size(used_memory(1))


def profiling(func):
    """
    Create the Decorator to mark a function for profiling. The time and memory usage will be recorded and printed.

    Example:
    .. highlight:: python
    .. code-block:: python

        @profiling
        def foo():
            print(1)

    :param func: function to be profiled
    :return: arguments wrapper
    """
    from . import default_logger

    @wraps(func)
    def arg_wrapper(*args, **kwargs):
        start_t = time.perf_counter()
        start_mem = used_memory(unit=1)
        r = func(*args, **kwargs)
        elapsed = time.perf_counter() - start_t
        end_mem = used_memory(unit=1)
        # level_prefix = ''.join('-' for v in inspect.stack() if v and v.index is not None and v.index >= 0)
        level_prefix = ''
        mem_status = f'memory Δ {get_readable_size(end_mem - start_mem)} {get_readable_size(start_mem)} -> {get_readable_size(end_mem)}'
        default_logger.info(
            f'{level_prefix} {func.__qualname__} time: {elapsed}s {mem_status}'
        )
        return r

    return arg_wrapper


class TimeDict:
    """Records of time information."""

    def __init__(self):
        self.accum_time = defaultdict(float)
        self.first_start_time = defaultdict(float)
        self.start_time = defaultdict(float)
        self.end_time = defaultdict(float)
        self._key_stack = []
        self._pending_reset = False

    def __enter__(self):
        _key = self._key_stack[-1]
        # store only the first enter time
        if _key not in self.first_start_time:
            self.first_start_time[_key] = time.perf_counter()
        self.start_time[_key] = time.perf_counter()
        return self

    def __exit__(self, typ, value, traceback):
        _key = self._key_stack.pop()
        self.end_time[_key] = time.perf_counter()
        self.accum_time[_key] += self.end_time[_key] - self.start_time[_key]
        if self._pending_reset:
            self.reset()

    def __call__(self, key: str, *args, **kwargs):
        """
        Add time counter.

        :param key: key name of the counter
        :param args: extra arguments
        :param kwargs: keyword arguments
        :return: self object
        """
        self._key_stack.append(key)
        return self

    def reset(self):
        """Clear the time information."""
        if self._key_stack:
            self._pending_reset = True
        else:
            self.accum_time.clear()
            self.start_time.clear()
            self.first_start_time.clear()
            self.end_time.clear()
            self._key_stack.clear()
            self._pending_reset = False

    def __str__(self):
        return ' '.join(f'{k}: {v:3.1f}s' for k, v in self.accum_time.items())


class TimeContext:
    """Timing a code snippet with a context manager."""

    time_attrs = ['years', 'months', 'days', 'hours', 'minutes', 'seconds']

    def __init__(self, task_name: str, logger: 'JinaLogger' = None):
        """
        Create the context manager to timing a code snippet.

        :param task_name: The context/message.
        :param logger: Use existing logger or use naive :func:`print`.

        Example:
        .. highlight:: python
        .. code-block:: python

            with TimeContext('loop'):
                do_busy()

        """
        self.task_name = task_name
        self._logger = logger
        self.duration = 0

    def __enter__(self):
        self.start = time.perf_counter()
        self._enter_msg()
        return self

    def _enter_msg(self):
        if self._logger:
            self._logger.info(self.task_name + '...')
        else:
            print(self.task_name, end=' ...\t', flush=True)

    def __exit__(self, typ, value, traceback):
        self.duration = self.now()

        self.readable_duration = get_readable_time(seconds=self.duration)

        self._exit_msg()

    def now(self) -> float:
        """
        Get the passed time from start to now.

        :return: passed time
        """
        return time.perf_counter() - self.start

    def _exit_msg(self):
        if self._logger:
            self._logger.info(
                f'{self.task_name} takes {self.readable_duration} ({self.duration:.2f}s)'
            )
        else:
            print(
                colored(
                    f'{self.task_name} takes {self.readable_duration} ({self.duration:.2f}s)'
                ),
                flush=True,
            )


class ProgressBar(TimeContext):
    """
    A simple progress bar.

    Example:
        .. highlight:: python
        .. code-block:: python

            with ProgressBar('loop'):
                do_busy()
    """

    def __init__(
        self,
        bar_len: int = 20,
        task_name: str = '',
        batch_unit: str = 'requests',
        logger=None,
    ):
        """
        Create the ProgressBar.

        :param bar_len: Total length of the bar.
        :param task_name: The name of the task, will be displayed in front of the bar.
        :param batch_unit: Unit of batch
        :param logger: Jina logger
        """
        super().__init__(task_name, logger)
        self.bar_len = bar_len
        self.num_docs = 0
        self._ticks = 0
        self.batch_unit = batch_unit

    def update_tick(self, tick: float = 0.1) -> None:
        """
        Increment the progress bar by one tick, when the ticks accumulate to one, trigger one :meth:`update`.

        :param tick: A float unit to increment (should < 1).
        """
        self._ticks += tick
        if self._ticks > 1:
            self.update()
            self._ticks = 0

    def update(self, progress: Optional[int] = None, *args, **kwargs) -> None:
        """
        Increment the progress bar by one unit.

        :param progress: The number of unit to increment.
        :param args: extra arguments
        :param kwargs: keyword arguments
        """
        self.num_reqs += 1
        sys.stdout.write('\r')
        elapsed = time.perf_counter() - self.start
        num_bars = self.num_reqs % self.bar_len
        num_bars = self.bar_len if not num_bars and self.num_reqs else max(num_bars, 1)
        if progress:
            self.num_docs += progress

        sys.stdout.write(
            '{:>10} |{:<{}}| 📃 {:6d} ⏱️ {:3.1f}s 🐎 {:3.1f}/s {:6d} {:>10}'.format(
                colored(self.task_name, 'cyan'),
                colored('█' * num_bars, 'green'),
                self.bar_len + 9,
                self.num_docs,
                elapsed,
                self.num_docs / elapsed,
                self.num_reqs,
                self.batch_unit,
            )
        )
        if num_bars == self.bar_len:
            sys.stdout.write('\n')
        sys.stdout.flush()
        from . import profile_logger

        profile_logger.info(
            {
                'num_bars': num_bars,
                'num_reqs': self.num_reqs,
                'bar_len': self.bar_len,
                'progress': num_bars / self.bar_len,
                'task_name': self.task_name,
                'qps': self.num_reqs / elapsed,
                'speed': (self.num_docs if self.num_docs > 0 else self.num_reqs)
                / elapsed,
                'speed_unit': ('Documents' if self.num_docs > 0 else 'Requests'),
                'elapsed': elapsed,
            }
        )

    def __enter__(self):
        super().__enter__()
        self.num_reqs = -1
        self.num_docs = 0
        self.update()
        return self

    def _enter_msg(self):
        pass

    def _exit_msg(self):
        if self.num_docs > 0:
            speed = self.num_docs / self.duration
        else:
            speed = self.num_reqs / self.duration
        sys.stdout.write(
            f'\t{colored(f"✅ done in ⏱ {self.readable_duration} 🐎 {speed:3.1f}/s", "green")}\n'
        )
